"""Test the MAST Engineering interface"""

import logging
import os
from pathlib import Path

import pytest
from astropy.table import Table
from astropy.time import Time
from astropy.utils.diff import report_diff_values

from romancal.lib.engdb import engdb_edp
from romancal.lib.engdb.engdb_lib import EngDB_Value

pytest.importorskip("edp.mnemonics_reader")

# Configure logging
log = logging.getLogger(__name__)

# Test query
QUERY = ("ope_scf_dir", "2027-02-23T01:00:00", "2027-02-23T01:00:05")

# Expected return from query
EXPECTED_RESPONSE = {
    "TlmMnemonic": "ope_scf_dir",
    "ReqSTime": "2027-02-23T01:00:00.000",
    "ReqETime": "2027-02-23T01:00:05.000",
    "Count": 7,
    "AllPoints": 1,
    "Data": [
        {"ObsTime": "2027-02-23T00:59:59.054", "EUValue": "SCFA"},
        {"ObsTime": "2027-02-23T01:00:00.052", "EUValue": "SCFA"},
        {"ObsTime": "2027-02-23T01:00:01.052", "EUValue": "SCFA"},
        {"ObsTime": "2027-02-23T01:00:02.162", "EUValue": "SCFA"},
        {"ObsTime": "2027-02-23T01:00:03.161", "EUValue": "SCFA"},
        {"ObsTime": "2027-02-23T01:00:04.161", "EUValue": "SCFA"},
        {"ObsTime": "2027-02-23T01:00:05.163", "EUValue": "SCFA"},
    ],
}
EXPECTED_RECORDS = Table.read(
    Path(__file__).parent / "data" / "test_records_expected.ecsv", format="ascii.ecsv"
)


def test_get_records(engdb):
    """Test getting records"""
    records = engdb._get_records(*QUERY)
    assert engdb.response == EXPECTED_RESPONSE
    assert report_diff_values(records, EXPECTED_RECORDS)


@pytest.mark.parametrize(
    "pars, expected",
    [
        ({}, ["SCFA"] * 5),
        (
            {"include_obstime": True},
            [
                EngDB_Value(
                    obstime=Time(61459.04166726852, format="mjd"), value="SCFA"
                ),
                EngDB_Value(
                    obstime=Time(61459.041678842594, format="mjd"), value="SCFA"
                ),
                EngDB_Value(
                    obstime=Time(61459.04169168982, format="mjd"), value="SCFA"
                ),
                EngDB_Value(
                    obstime=Time(61459.04170325232, format="mjd"), value="SCFA"
                ),
                EngDB_Value(
                    obstime=Time(61459.04171482639, format="mjd"), value="SCFA"
                ),
            ],
        ),
        (
            {"include_obstime": True, "zip_results": False},
            EngDB_Value(
                obstime=[
                    Time(61459.04166726852, format="mjd"),
                    Time(61459.041678842594, format="mjd"),
                    Time(61459.04169168982, format="mjd"),
                    Time(61459.04170325232, format="mjd"),
                    Time(61459.04171482639, format="mjd"),
                ],
                value=["SCFA"] * 5,
            ),
        ),
        ({"include_bracket_values": True}, ["SCFA"] * 7),
    ],
)
def test_get_values(engdb, pars, expected):
    values = engdb.get_values(*QUERY, **pars)
    assert values == expected


@pytest.fixture(scope="module")
def engdb():
    """Open a connection"""
    try:
        engdb = engdb_edp.EngdbEDP("test", os.environ.get("KRB5CCPATH", None))
    except RuntimeError as exception:
        pytest.skip(f"Live Engineering Service not available: {exception}")
    return engdb



import numpy as np
from drizzle.resample import Drizzle
from roman_datamodels import dqflags
from stcal.resample.utils import (
    build_driz_weight,
)


class ExptimeResampler:
    def __init__(self, out_wcs, out_shape, good_bits, kernel):
        self.out_wcs = out_wcs
        self.out_shape = out_shape
        self.good_bits = good_bits

        self.out_img = np.zeros(out_shape, dtype="f4")
        self.out_wht = np.zeros(out_shape, dtype="f4")
        self.exptime_total = np.zeros(out_shape, dtype="f4")

        self._driz = Drizzle(
            kernel=kernel,
            fillval=0,
            out_img=self.out_img,
            out_wht=self.out_wht,
            disable_ctx=True,
        )

    def add_image(self, model, pixmap, xmin, xmax, ymin, ymax):
        data = np.full(model["data"].shape, model["effective_exposure_time"])

        # create a unit weight map for all the input pixels with science data
        inwht = build_driz_weight(
            {
                "data": model["data"],
                "dq": model["dq"],
            },
            weight_type=None,
            good_bits=self.good_bits,
            flag_name_map=dqflags.pixel,
        )

        self._driz.add_image(
            data,
            pixmap=pixmap,
            exptime=1.0,
            scale=1.0,
            weight_map=inwht,
            wht_scale=1.0,
            pixfrac=1.0,
            in_units="cps",
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
        )

        self.exptime_total += self.out_img

        # zero out arrays for next image
        self.out_img[:] = 0
        self.out_wht[:] = 0

    def finalize(self):
        return self.exptime_total



#!/usr/bin/env python

"""
Set the initial world coordinate system for Roman exposures.

The Roman engineering database is queried for the Roman Space Telescope observatory
orientation parameters, and converts that orientation to a WCS
for a list of exposures.
"""

# Licensed under a 3-clause BSD style license - see LICENSE

import argparse
import logging
import warnings
from pathlib import Path

import romancal.orientation.set_telescope_pointing as stp
from romancal.lib.engdb.engdb_tools import AVAILABLE_SERVICES


def main():
    """Set the initial world coordinate system."""
    parser = argparse.ArgumentParser(
        description=(
            "Update basic WCS information in Roman exposures from the engineering database."
            " For detailed information, see"
            " https://roman-pipeline.readthedocs.io/en/latest/roman/orientation/set_telescope_pointing.html"
        )
    )
    parser.add_argument(
        "exposure", type=str, nargs="+", help="List of Roman exposures to update."
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase verbosity. Specifying multiple times adds more output.",
    )
    parser.add_argument(
        "--allow-default",
        action="store_true",
        help="If pointing information cannot be determine, use header information.",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Perform all actions but do not save the results",
    )
    parser.add_argument(
        "--save-transforms", action="store_true", help="Save transforms."
    )
    parser.add_argument(
        "--tolerance",
        type=int,
        default=60,
        help="Seconds beyond the observation time to search for telemetry. Default: %(default)s",
    )
    parser.add_argument(
        "--service",
        type=str,
        default="mast",
        choices=[name for name in AVAILABLE_SERVICES],
        help="Database service to use. Default: %(default)s",
    )

    # Arguments pertinent only to the EngdbMast service.
    if "mast" in AVAILABLE_SERVICES:
        parser.add_argument(
            "--engdb-url",
            type=str,
            default=None,
            help=(
                "URL of the engineering database."
                " If not specified, the environment variable 'ENG_BASE_URL' is used."
                " Otherwise, a hardwired default is used."
            ),
        )

    # Arguments pertinent only to the EngdbEDP service
    if "edp" in AVAILABLE_SERVICES:
        parser.add_argument(
            "--environment",
            type=str,
            default="test",
            choices=["dev", "test", "int", "ops"],
            help="Operational environment in use. Default: %(default)s",
        )
        parser.add_argument(
            "--path-to-cc",
            type=str,
            help="Full path to the required kerberos authentication keytab file",
        )

    args = parser.parse_args()

    # Set output detail.
    logger = logging.getLogger("romancal")
    logger_handler = logging.StreamHandler()
    logger.addHandler(logger_handler)
    logger_format_debug = logging.Formatter(
        "%(levelname)s:%(filename)s::%(funcName)s: %(message)s"
    )
    level = stp.LOGLEVELS[min(len(stp.LOGLEVELS) - 1, args.verbose)]
    logger.setLevel(level)
    if level <= logging.DEBUG:
        logger_handler.setFormatter(logger_format_debug)
    logger.info("set_telescope_pointing called with args %s", args)

    # Gather the service-specific args
    service_kwargs = {"service": args.service}
    for arg in ["engdb_url", "environment", "path_to_cc"]:
        try:
            service_kwargs[arg] = getattr(args, arg)
        except AttributeError:
            pass

    # Calculate WCS for all inputs.
    for filename in args.exposure:
        logger.info("")
        logger.info("------")
        logger.info(f"Setting pointing for {filename}")

        # Create path for saving the transforms.
        transform_path = None
        if args.save_transforms:
            path = Path(filename)
            transform_path = path.with_name(f"{path.stem}_transforms.asdf")

        try:
            stp.add_wcs(
                filename,
                tolerance=args.tolerance,
                allow_default=args.allow_default,
                dry_run=args.dry_run,
                save_transforms=transform_path,
                service_kwargs=service_kwargs,
            )
        except (TypeError, ValueError) as exception:
            logger.warning("Cannot determine pointing information: %s", str(exception))
            logger.debug("Full exception:", exc_info=exception)


def deprecated_name():
    """Raise warning if filename.* is no longer used, and provide correct one."""
    filename = Path(__file__)
    warnings.warn(
        f"usage of `{filename.name}` is deprecated; use `{filename.stem}` instead",
        stacklevel=2,
    )

    main()


if __name__ == "__main__":
    main()



"""Unit tests for dark current correction"""

import numpy as np
import pytest
import roman_datamodels as rdm
from roman_datamodels import stnode
from roman_datamodels.datamodels import DarkRefModel, ImageModel

from romancal.dark_current import DarkCurrentStep


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ],
)
def test_dark_step_interface(instrument, exptype):
    """Test that the basic inferface works for data requiring a DARK reffile"""
    # Set test size, data + reference files
    ref_shape = (10, 28, 28)
    image_shape = (ref_shape[1], ref_shape[2])

    # Create test rampfit and dark models
    rampfit_model, darkref_model = create_image_and_dark(ref_shape, instrument, exptype)
    # Perform Dark Current subtraction step
    result = DarkCurrentStep.call(rampfit_model, override_dark=darkref_model)

    # Test dark results
    trim_shape = (image_shape[0] - 8, image_shape[1] - 8)  # size of the trimmed image
    assert (result.data == rampfit_model.data).all()
    assert isinstance(result, ImageModel)
    assert result.validate() is None
    assert result.data.shape == trim_shape
    assert result.dq.shape == trim_shape
    assert result.meta.cal_step.dark == "COMPLETE"
    assert result.data.dtype == np.float32
    assert result.dq.dtype == np.uint32


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ],
)
def test_dark_step_subtraction(instrument, exptype):
    """Test that the values in a dark reference file are properly subtracted"""
    # Set test size
    ref_shape = (10, 28, 28)

    # Create test ramp and dark models
    image_model, darkref_model = create_image_and_dark(ref_shape, instrument, exptype)

    # populate data array of science cube
    orig_model = image_model.copy()

    # Perform Dark Current subtraction step
    result = DarkCurrentStep.call(image_model, override_dark=darkref_model)

    diff = orig_model.data - (darkref_model.dark_slope[4:-4, 4:-4])

    # test that the output data file is equal to the difference found when subtracting
    # reffile from sci file
    assert np.all(np.isclose(result.data, diff, atol=1.0e-7))


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ],
)
def test_dark_step_output_dark_file(tmp_path, instrument, exptype):
    """Test that the the step can output a proper (optional) dark file"""
    path = str(tmp_path / "dark_out.asdf")

    # Set test size
    ref_shape = (10, 28, 28)
    image_shape = (ref_shape[1], ref_shape[2])

    # Create test ramp and dark models
    image_model, darkref_model = create_image_and_dark(ref_shape, instrument, exptype)

    # Perform Dark Current subtraction step
    DarkCurrentStep.call(image_model, override_dark=darkref_model, dark_output=path)

    # Open dark file
    dark_out_file_model = rdm.open(path)

    # Test dark file results
    assert isinstance(dark_out_file_model, DarkRefModel)
    assert dark_out_file_model.validate() is None
    assert dark_out_file_model.dq.shape == image_shape


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ],
)
def test_dark_step_getbestrefs(tmp_path, instrument, exptype):
    """Test that the the step will skip if CRDS returns N/A for the ref file"""
    path = str(tmp_path / "dark_out.asdf")

    # Set test size
    ref_shape = (10, 28, 28)
    image_shape = (ref_shape[1], ref_shape[2])

    # Create test ramp and dark models
    image_model, darkref_model = create_image_and_dark(ref_shape, instrument, exptype)

    # Perform Dark Current subtraction step with override = N/A
    result = DarkCurrentStep.call(image_model, override_dark="N/A")
    assert result.meta.cal_step.dark == "SKIPPED"

    # Perform Dark Current subtraction step
    DarkCurrentStep.call(image_model, override_dark=darkref_model, dark_output=path)

    # Open dark file
    dark_out_file_model = rdm.open(path)

    # Test dark file results
    assert isinstance(dark_out_file_model, DarkRefModel)
    assert dark_out_file_model.validate() is None
    assert dark_out_file_model.dq.shape == image_shape


def create_image_and_dark(shape, instrument, exptype):
    """Helper function to create test image and dark models"""
    # Create test image model
    image_shape = (shape[1], shape[2])
    image = ImageModel.create_fake_data(shape=image_shape)
    image.meta.cal_step = stnode.L2CalStep.create_fake_data()
    image.meta.cal_logs = stnode.CalLogs.create_fake_data()
    image.meta.instrument.name = instrument
    image.meta.instrument.detector = "WFI01"
    image.meta.instrument.optical_element = "F158"
    image.meta.exposure.type = exptype
    image.meta.exposure.read_pattern = [[1], [2, 3], [4], [5, 6, 7, 8], [9, 10], [11]]
    image.data = np.ones(image_shape, dtype=np.float32)
    image.data = np.full(image_shape, 0.2, dtype=np.float32)
    image.dq = np.zeros(image_shape, dtype=image.dq.dtype)
    # Trim the image
    image.data = image.data[4:-4, 4:-4]
    image.dq = image.dq[4:-4, 4:-4]
    image.err = image.err[4:-4, 4:-4]

    # Create dark model
    darkref = DarkRefModel.create_fake_data(shape=shape)
    darkref.dark_slope = np.full((image_shape), 5.3e-03, dtype=np.float32)
    darkref.dark_slope_error = np.full((image_shape), 2.6e-05, dtype=np.float32)
    darkref.dq = np.zeros(image_shape, dtype=image.dq.dtype)

    return image, darkref



"""
Test suite for set_telescope_pointing
"""

import dataclasses
import logging
from pathlib import Path

import numpy as np
import pytest

pytest.importorskip("pysiaf")

import asdf
import roman_datamodels as rdm
from astropy.time import Time

from romancal.lib.engdb import engdb_mast
from romancal.orientation import set_telescope_pointing as stp

# Ensure that `set_telescope_pointing` logs.
stp.logger.setLevel(logging.DEBUG)
stp.logger.addHandler(logging.StreamHandler())

# Setup for some times
STARTTIME = Time("2027-03-23T19:20:40", format="isot")
ENDTIME = Time("2027-03-23T19:21:36", format="isot")
BADSTARTTIME = Time("2020-02-02T02:02:02", format="isot")
BADENDTIME = Time("2020-02-02T02:12:02", format="isot")

# Header defaults
TARG_RA = 270.0
TARG_DEC = 66.01

# Get the mock databases
DATA_PATH = Path(__file__).parent / "data"

OBSTIME_EXPECTED = Time(1805829668.0276294, format="unix")

# Meta attributes for test comparisons
METAS_EQUALITY = [
    "meta.pointing.pointing_engineering_source",
]
METAS_ISCLOSE = [
    "meta.wcsinfo.dec_ref",
    "meta.wcsinfo.ra_ref",
    "meta.wcsinfo.roll_ref",
    "meta.wcsinfo.v2_ref",
    "meta.wcsinfo.v3_ref",
    "meta.pointing.ra_v1",
    "meta.pointing.dec_v1",
    "meta.pointing.pa_v3",
]


def test_add_wcs_default(science_raw_model, tmp_path):
    """Handle when no pointing exists and the default is used."""
    m = science_raw_model
    m.meta.exposure.start_time = Time("2022-01-01T00:00:00")
    m.meta.exposure.end_time = Time("2022-01-01T01:00:00")
    model_path = _model_to_tmpfile(m, tmp_path)
    stp.add_wcs(model_path, tolerance=0, allow_default=True)

    with rdm.open(model_path) as result:
        assert result.meta.wcsinfo.ra_ref == result.meta.pointing.target_ra
        assert result.meta.wcsinfo.dec_ref == result.meta.pointing.target_dec


def test_change_base_url():
    """Test changing the engineering database by call for success.

    The given time and database should not find any values.
    """
    service_kwargs = {"service": "mast", "base_url": engdb_mast.MAST_BASE_URL}
    with pytest.raises(ValueError):
        stp.get_pointing(
            Time("2015-06-15"), Time("2015-06-17"), service_kwargs=service_kwargs
        )


def test_change_base_url_fail():
    """Test changing the engineering database by call"""
    service_kwargs = {"service": "mast", "base_url": "https://nonexistent.fake.example"}
    with pytest.raises(ValueError):
        stp.get_pointing(
            Time(STARTTIME, format="isot"),
            Time(ENDTIME, format="isot"),
            service_kwargs=service_kwargs,
        )


def test_get_pointing():
    """Ensure that the averaging works."""
    q_expected = np.array([-0.52558752, 0.3719724, -0.52016581, 0.38150882])
    obstime, q = stp.get_pointing(STARTTIME, ENDTIME)

    assert np.isclose(obstime.value, OBSTIME_EXPECTED.value)
    assert np.allclose(q, q_expected)


def test_get_pointing_fail():
    with pytest.raises(ValueError):
        obstime, q = stp.get_pointing(BADSTARTTIME, BADENDTIME)


def test_get_pointing_list():
    q_expected = np.array([-0.690189, 0.121953, -0.695103, 0.159999])
    results = stp.get_pointing(STARTTIME, ENDTIME, reduce_func=stp.all_pointings)
    assert isinstance(results, list)
    assert len(results) > 0
    assert np.isclose(results[0].q, q_expected).all()
    assert STARTTIME <= results[0].obstime <= ENDTIME


def test_logging(caplog):
    stp.get_pointing(STARTTIME, ENDTIME)
    assert "Determining pointing between observations times" in caplog.text
    assert "Telemetry search tolerance" in caplog.text
    assert "Reduction function" in caplog.text
    assert "Querying engineering DB" in caplog.text


@pytest.mark.parametrize("wcs_type", ["wcsinfo", "vinfo"])
def test_wcs(calc_wcs, wcs_type):
    """Ensure WCS information is correct

    Parameters
    ----------
    calc_wcs : pytest.fixture
        Equates to tuple (wcsinfo, vinfo, transforms, t_pars)

    wcs_type : str
        Which particular WCS information to test
    """
    wcsinfo, vinfo, _, t_pars = calc_wcs
    wcs = {"wcsinfo": wcsinfo, "vinfo": vinfo}
    with asdf.open(DATA_PATH / "wcs.asdf") as af:
        expected = af.tree[wcs_type]

    wcs_dict = wcs[wcs_type]._asdict()
    for key in wcs_dict:
        if key != "s_region":
            assert np.isclose(expected[key], wcs_dict[key]), (
                f"Key {key} differs expected {expected[key]} calculated {wcs_dict[key]}"
            )
        else:
            assert expected[key] == wcs_dict[key]


def test_strict_pointing(science_raw_model, tmp_path):
    """Test failure on strict pointing"""
    science_raw_model.meta.exposure.end_time = STARTTIME
    with pytest.raises(ValueError):
        stp.add_wcs(_model_to_tmpfile(science_raw_model, tmp_path), tolerance=0)


@pytest.mark.parametrize(
    "matrix", [matrix for matrix in dataclasses.fields(stp.Transforms())]
)
def test_transforms(calc_wcs, matrix):
    """Ensure expected calculate of the specified matrix

    Parameters
    ----------
    transforms, t_pars : Transforms, TransformParameters
        The transforms and the parameters used to generate the transforms

    matrix : str
        The matrix to compare
    """
    _, _, transforms, t_pars = calc_wcs
    _test_transforms(transforms, t_pars, matrix.name)


def test_transform_serialize(calc_wcs, tmp_path):
    """Test serialization of Transforms"""
    _, _, transforms, _ = calc_wcs

    path = tmp_path / "transforms.asdf"
    transforms.write_to_asdf(path)
    from_asdf = stp.Transforms.from_asdf(path)

    assert isinstance(from_asdf, stp.Transforms)
    assert str(transforms) == str(from_asdf)


# ######################
# Utilities and fixtures
# ######################
@pytest.fixture
def calc_wcs(tmp_path_factory):
    """Calculate full transforms and WCS info"""
    t_pars = _make_t_pars()

    # Calculate the transforms and WCS information
    wcsinfo, vinfo, transforms = stp.calc_wcs(t_pars)

    # Save all for later examination.
    transforms_path = tmp_path_factory.mktemp("transforms")
    transforms.write_to_asdf(transforms_path / "transforms.asdf")
    wcs_asdf_file = asdf.AsdfFile(
        {"wcsinfo": wcsinfo._asdict(), "vinfo": vinfo._asdict()}
    )
    wcs_asdf_file.write_to(transforms_path / "wcs.asdf")

    return wcsinfo, vinfo, transforms, t_pars


@pytest.fixture
def science_raw_model():
    """Create the base model for testing"""
    m = rdm.datamodels.ScienceRawModel.create_fake_data(
        {
            "meta": {
                "exposure": {"start_time": STARTTIME, "end_time": ENDTIME},
                "pointing": {"target_ra": TARG_RA, "target_dec": TARG_DEC},
                "wcsinfo": {"aperture_name": "WFI_CEN"},
            }
        }
    )

    return m


def _make_t_pars(aperture="WFI02_FULL"):
    """Setup initial Transforms Parameters

    This set is Visit 1 provided by T.Sohn in a demonstration notebook.

    Parameters
    ==========
    aperture : str
        Aperture in use.
    """
    t_pars = stp.TransformParameters()

    t_pars.aperture = aperture

    t_pars.pointing = stp.Pointing(
        q=np.array([0.709433, -0.291814, 0.641319, 0.016107]),
    )

    t_pars.service_kwargs = {"service": "mast"}

    return t_pars


def _model_to_tmpfile(m, tmp_path, fname="file.asdf"):
    """Save a DataModel to a general temp file"""
    file_path = tmp_path / fname
    m.save(file_path)
    m.close()
    return file_path


def _test_transforms(transforms, t_pars, matrix):
    """Private function to ensure expected calculate of the specified matrix

    Parameters
    ----------
    transforms, t_pars : Transforms, TransformParameters
        The transforms and the parameters used to generate the transforms

    matrix : str
        The matrix to compare
    """
    expected_tforms = stp.Transforms.from_asdf(DATA_PATH / "transforms.asdf")
    expected_value = getattr(expected_tforms, matrix)

    value = getattr(transforms, matrix)
    if expected_value is None:
        assert value is None
    else:
        assert np.allclose(value, expected_value)



#!/usr/bin/env python

"""Add velocity aberration correction information to the FITS files provided."""

import argparse
import sys

from roman_datamodels import open as dm_open
from stcal.velocity_aberration import compute_va_effects

__all__ = []  # type: ignore[var-annotated]


def parse_args(args):
    """
    Parse given arguments.

    Parameters
    ----------
    args : command-line input
        The arguments to parse

    Returns
    -------
    parser : object
        Parsed arguments
    """
    description_text = """
    Add velocity aberration correction information to Roman data.
    Input via command line can be one or more files.
    It assumes the following meta keys are present:
    meta.ephemeris.velocity_x (km/sec),
    meta.ephemeris.velocity_y (km/sec),
    meta.ephemeris.velocity_z (km/sec),
    meta.wcsinfo.dec_ref (deg),
    meta.wcsinfo.ra_ref (deg),

    The meta keys updates are:
    meta.velocity_aberration.dec_reference (deg)
    meta.velocity_aberration.ra_reference (deg)
    meta.velocity_aberration.scale_factor (unitless multiplication factor)
    """
    parser = argparse.ArgumentParser(
        prog="set_velocity_aberration",
        description=description_text,
    )
    parser.add_argument(
        "filename",
        nargs="+",
        help="Roman file(s) to which to add the velocity aberration correction information.",
    )
    return parser.parse_args(args)


def main():
    """Parse arguments and add velocity aberration correction information to the files provided."""
    args = parse_args(sys.argv[1:])
    for filename in args.filename:
        add_dva(filename)


def add_dva(filename):
    """
    Determine velocity aberration.

    Given the name of a valid partially populated level ScienceRawModel file,
    determine the velocity aberration scale factor and apparent target position
    in the moving (telescope) frame.

    It presumes all the accessed keywords are present (see first block).

    Parameters
    ----------
    filename : str
        The name of the file to be updated.
    """
    model = dm_open(filename)
    scale_factor, apparent_ra, apparent_dec = compute_va_effects(
        velocity_x=model.meta.ephemeris.velocity_x,
        velocity_y=model.meta.ephemeris.velocity_y,
        velocity_z=model.meta.ephemeris.velocity_z,
        ra=model.meta.wcsinfo.ra_ref,
        dec=model.meta.wcsinfo.dec_ref,
    )

    # update header
    model.meta.velocity_aberration.scale_factor = scale_factor
    model.meta.velocity_aberration.ra_reference = apparent_ra
    model.meta.velocity_aberration.dec_reference = apparent_dec
    model.save(filename)


if __name__ == "__main__":
    main()



import json
import logging

import pytest
from astropy.time import Time
from roman_datamodels.datamodels import FlatRefModel, ImageModel
from stpipe import crds_client

import romancal
from romancal.datamodels import ModelLibrary
from romancal.flatfield import FlatFieldStep
from romancal.stpipe import RomanPipeline, RomanStep


@pytest.mark.parametrize("is_container", [True, False])
@pytest.mark.parametrize("step_class", [RomanPipeline, RomanStep])
def test_open_model(step_class, tmp_path, is_container, base_image):
    """
    Test that the class is properly hooked up to datamodels.open.
    More comprehensive tests can be found in romancal.datamodels.tests,
    this is just a smoke test of the integration.
    """
    file_path = tmp_path / "test.asdf"

    im = base_image()
    im.save(file_path)

    if is_container:
        asn = {
            "asn_pool": "none",
            "products": [
                {
                    "members": [
                        {
                            "exptype": "science",
                            "expname": "test.asdf",
                        }
                    ],
                },
            ],
        }
        asn_path = tmp_path / "test.json"
        with open(asn_path, "w") as f:
            json.dump(asn, f)
        test_file_path = asn_path
    else:
        test_file_path = file_path

    step = step_class()
    with step.open_model(test_file_path) as model:
        if is_container:
            assert isinstance(model, ModelLibrary)
            assert model.crds_observatory == "roman"
            assert model.get_crds_parameters() is not None
        else:
            assert isinstance(model, ImageModel)
            assert model.meta.telescope == "ROMAN"


@pytest.mark.parametrize("step_class", [RomanPipeline, RomanStep])
def test_get_reference_file(step_class, base_image):
    """
    Test that CRDS is properly integrated.
    """
    im = base_image()
    # This will be brittle while we're using the dev server.
    # If this test starts failing mysteriously, check the
    # metadata values against the flat rmap.
    im.meta.instrument.optical_element = "F158"
    im.meta.exposure.start_time = Time("2021-01-01T12:00:00")

    step = step_class()
    reference_path = step.get_reference_file(im, "flat")

    with step.open_model(reference_path) as reference_model:
        assert isinstance(reference_model, FlatRefModel)


@pytest.mark.skip(reason="There are no grism flats.")
@pytest.mark.parametrize("step_class", [RomanPipeline, RomanStep])
def test_get_reference_file_spectral(step_class, base_image):
    """
    Test that CRDS is properly integrated.
    """
    im = base_image()
    # This will be brittle while we're using the dev server.
    # If this test starts failing mysteriously, check the
    # metadata values against the flat rmap.
    im.meta.instrument.optical_element = "GRISM"
    im.meta.exposure.start_time = Time("2021-01-01T12:00:00")

    step = step_class()
    reference_path = step.get_reference_file(im, "flat")

    with step.open_model(reference_path) as reference_model:
        assert isinstance(reference_model, FlatRefModel)
        assert reference_model.meta.instrument.optical_element == "GRISM"


def test_log_messages(tmp_path, base_image):
    logger = logging.getLogger("test_log_messages")

    class LoggingStep(RomanStep):
        def process(self):
            logger.warning("Splines failed to reticulate")
            return base_image()

    result = LoggingStep().run()
    assert any("Splines failed to reticulate" in l for l in result.meta.cal_logs)


def test_crds_meta(base_image):
    """Test that context and software versions are set"""

    im = base_image()
    result = FlatFieldStep.call(im)

    assert result.meta.ref_file.crds.version == crds_client.get_svn_version()
    assert result.meta.ref_file.crds.context == crds_client.get_context_used(
        result.crds_observatory
    )


def test_calibration_software_version(base_image):
    """Test that calibration_software_version is updated when a step is run"""

    class NullStep(RomanStep):
        def process(self, input):
            return input

    im = base_image()
    im.meta.calibration_software_version = "junkversion"

    result = NullStep.call(im)

    assert result.meta.calibration_software_version == romancal.__version__



from pathlib import Path

import numpy as np
import pytest
import roman_datamodels as rdm
from gwcs.wcstools import grid_from_bounding_box
from numpy.testing import assert_allclose
from roman_datamodels.datamodels import ImageModel
from roman_datamodels.dqflags import pixel

from romancal.assign_wcs.assign_wcs_step import AssignWcsStep
from romancal.lib.suffix import replace_suffix
from romancal.pipeline.exposure_pipeline import ExposurePipeline

from .regtestdata import compare_asdf

# mark all tests in this module
pytestmark = pytest.mark.bigdata


@pytest.fixture(scope="module")
def run_elp(rtdata_module, resource_tracker):
    rtdata = rtdata_module

    input_data = "r0000101001001001001_0001_wfi01_f158_uncal.asdf"
    rtdata.get_data(f"WFI/image/{input_data}")
    rtdata.input = input_data

    # Test Pipeline
    output = "r0000101001001001001_0001_wfi01_f158_cal.asdf"
    rtdata.output = output
    args = [
        "roman_elp",
        rtdata.input,
    ]
    with resource_tracker.track():
        ExposurePipeline.from_cmdline(args)

    # get truth file
    rtdata.get_truth(f"truth/WFI/image/{output}")
    return rtdata


@pytest.fixture(scope="module")
def output_filename(run_elp):
    return run_elp.output


@pytest.fixture(scope="module")
def output_model(output_filename):
    with rdm.open(output_filename) as model:
        yield model


@pytest.fixture(scope="module")
def truth_filename(run_elp):
    return run_elp.truth


@pytest.fixture(scope="module")
def repointed_filename_and_delta(output_filename):
    delta = [10.0, 10.0]

    with rdm.open(output_filename) as model:
        # open from the file since we're going to modify the model
        del model.meta["wcs"]

        # Create new pointing for the model
        # RA & Dec are each shifted + 10 degrees, unless they are near
        # the upper limit, in which case they are shifted -10 degrees.
        if model.meta.wcsinfo.ra_ref >= 350.0:
            delta[0] *= -1.0
        if model.meta.wcsinfo.dec_ref >= 80.0:
            delta[1] *= -1.0

        model.meta.wcsinfo.ra_ref += delta[0]
        model.meta.wcsinfo.dec_ref += delta[1]

        # Create new wcs object for the new pointing
        model = AssignWcsStep.call(model)

        # save repointed model
        repointed_filename = output_filename.rsplit(".", 1)[0] + "_repoint.asdf"
        model.to_asdf(repointed_filename)

    return repointed_filename, delta


def test_log_tracked_resources(log_tracked_resources, run_elp):
    log_tracked_resources()


@pytest.mark.soctests
def test_output_matches_truth(output_filename, truth_filename, ignore_asdf_paths):
    diff = compare_asdf(output_filename, truth_filename, **ignore_asdf_paths)
    assert diff.identical, diff.report()


@pytest.mark.soctests
def test_output_is_image_model(output_model):
    # DMS280 result is an ImageModel
    assert isinstance(output_model, rdm.datamodels.ImageModel)


def test_wcs_matches_truth(run_elp, ignore_asdf_paths):
    output_path = Path(run_elp.output)
    wcs_filename = replace_suffix(output_path.stem, "wcs") + output_path.suffix
    wcs_path = output_path.parent / wcs_filename

    run_elp.output = wcs_filename
    run_elp.get_truth(f"truth/WFI/image/{wcs_filename}")

    truth_path = Path(run_elp.truth)

    diff = compare_asdf(wcs_path, truth_path, **ignore_asdf_paths)
    assert diff.identical, diff.report()


@pytest.mark.soctests
@pytest.mark.parametrize(
    "step_name",
    (
        "assign_wcs",
        "flat_field",
        "dark",
        "dq_init",
        "linearity",
        "ramp_fit",
        "saturation",
    ),
)
def test_steps_ran(output_model, step_name):
    # DMS86
    # also DMS129 for assign_wcs
    assert getattr(output_model.meta.cal_step, step_name) == "COMPLETE"


@pytest.mark.soctests
def test_jump_in_uneven_ramp(output_model):
    # DMS361 jump detection detected jumps in uneven ramp
    uneven = len({len(x) for x in output_model.meta.exposure.read_pattern}) > 1
    assert uneven & np.any(output_model.dq & pixel.JUMP_DET)


@pytest.mark.soctests
def test_has_a_wcs(output_model):
    # DMS129
    assert output_model.meta.wcs is not None


@pytest.mark.soctests
def test_wcs_has_distortion_information(output_model):
    # DMS129
    assert "v2v3" in output_model.meta.wcs.available_frames


@pytest.mark.soctests
def test_wcs_applies_distortion_correction(output_model):
    # DMS129
    # compare coordinates before and after distortion correction has been applied
    # 1 - get new image array based on the model
    x0, y0 = grid_from_bounding_box(output_model.meta.wcs.bounding_box)
    # 2 - apply the distortion-corrected WCS solution to new image array
    corrected_coords = output_model.meta.wcs(x0, y0)
    # 3 - apply the transformation from 'v2v3' to 'world' without distortion correction
    original_coords = output_model.meta.wcs.get_transform("v2v3", "world")(x0, y0)
    # compare both results to make sure they don't match
    # (which means the distortion correction was actually applied to the model)
    assert (corrected_coords[0] != original_coords[0]).all()
    assert (corrected_coords[1] != original_coords[1]).all()


@pytest.mark.soctests
@pytest.mark.parametrize("arr_name", ("dq", "err", "var_poisson", "var_rnoise"))
def test_array_exists(output_model, arr_name):
    # DMS87
    assert hasattr(output_model, arr_name)


@pytest.mark.soctests
def test_has_exposure_time(output_model):
    # DMS88 total exposure time exists
    assert "exposure_time" in output_model.meta.exposure


@pytest.mark.soctests
@pytest.mark.parametrize("meta_attribute", ("detector", "optical_element"))
def test_instrument_meta(output_model, meta_attribute):
    # DMS136 PSF tests
    assert meta_attribute in output_model.meta.instrument


@pytest.mark.soctests
def test_wcs_has_bounding_box(output_model):
    # DMS89 WCS tests
    assert len(output_model.meta.wcs.bounding_box) == 2


@pytest.mark.soctests
def test_repointed_matches_truth(
    repointed_filename_and_delta, rtdata, ignore_asdf_paths
):
    # DMS89 WCS tests
    repointed_filename, _ = repointed_filename_and_delta

    rtdata.get_truth(f"truth/WFI/image/{Path(repointed_filename).name}")
    rtdata.output = repointed_filename
    diff = compare_asdf(repointed_filename, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()


@pytest.mark.soctests
def test_repointed_wcs_differs(repointed_filename_and_delta, output_model):
    # DMS89 WCS tests
    repointed_filename, delta = repointed_filename_and_delta
    orig_wcs = output_model.meta.wcs
    with rdm.open(repointed_filename) as repointed_model:
        assert_allclose(
            [orig_wcs(2048, 2048)[0] + delta[0], orig_wcs(2048, 2048)[1] + delta[1]],
            repointed_model.meta.wcs(2048, 2048),
            atol=1.0,
        )


def test_elp_input_dm(rtdata, ignore_asdf_paths):
    """Test for input roman Datamodel to exposure level pipeline"""
    input_data = "r0000101001001001001_0001_wfi01_f158_uncal.asdf"
    rtdata.get_data(f"WFI/image/{input_data}")
    dm_input = rdm.open(rtdata.input)

    # Test Pipeline with input datamodel
    output = "r0000101001001001001_0001_wfi01_f158_cal.asdf"
    rtdata.output = output
    ExposurePipeline.call(dm_input, save_results=True)
    rtdata.get_truth(f"truth/WFI/image/{output}")

    # Ensure step completion is as expected
    with rdm.open(rtdata.output) as model:
        assert model.meta.cal_step.dq_init == "COMPLETE"
        assert model.meta.cal_step.saturation == "COMPLETE"
        assert model.meta.cal_step.linearity == "COMPLETE"
        assert model.meta.cal_step.dark == "COMPLETE"
        assert model.meta.cal_step.ramp_fit == "COMPLETE"
        assert model.meta.cal_step.assign_wcs == "COMPLETE"
        assert model.meta.cal_step.flat_field == "COMPLETE"
        assert model.meta.cal_step.photom == "COMPLETE"


@pytest.fixture(scope="module")
def run_all_saturated(rtdata_module):
    """
    Test ELP handling of an all saturated input model
    """
    rtdata = rtdata_module

    input_data = "r0000101001001001001_0001_wfi01_f158_ALL_SATURATED_uncal.asdf"
    rtdata.get_data(f"WFI/image/{input_data}")
    rtdata.input = input_data

    # Test Pipeline
    output = "r0000101001001001001_0001_wfi01_f158_ALL_SATURATED_cal.asdf"
    rtdata.output = output
    args = [
        "roman_elp",
        rtdata.input,
    ]
    ExposurePipeline.from_cmdline(args)

    # get truth file
    rtdata.get_truth(f"truth/WFI/image/{output}")
    return rtdata


@pytest.fixture(scope="module")
def all_saturated_model(run_all_saturated):
    with rdm.open(run_all_saturated.output) as model:
        yield model


def test_all_saturated_against_truth(run_all_saturated, ignore_asdf_paths):
    diff = compare_asdf(
        run_all_saturated.output, run_all_saturated.truth, **ignore_asdf_paths
    )
    assert diff.identical, diff.report()


@pytest.mark.parametrize(
    "step_name, status",
    [
        ("dq_init", "COMPLETE"),
        ("saturation", "COMPLETE"),
        ("linearity", "SKIPPED"),
        ("dark", "SKIPPED"),
        ("ramp_fit", "SKIPPED"),
        ("assign_wcs", "SKIPPED"),
        ("flat_field", "SKIPPED"),
        ("photom", "SKIPPED"),
        ("source_catalog", "SKIPPED"),
        ("tweakreg", "SKIPPED"),
    ],
)
def test_all_saturated_status(all_saturated_model, step_name, status):
    """
    For an all saturated input the pipeline should skip all steps after saturation.
    """
    assert getattr(all_saturated_model.meta.cal_step, step_name) == status


def test_all_saturated_model_type(all_saturated_model):
    """
    For an all saturated input the output model should be an ImageModel.
    """
    assert isinstance(all_saturated_model, ImageModel)


@pytest.mark.parametrize("array_name", ["data", "err", "var_poisson", "var_rnoise"])
def test_all_saturated_zeroed(all_saturated_model, array_name):
    """
    For an all saturated input the output model should contain 0s for data and err arrays.
    """
    np.testing.assert_array_equal(getattr(all_saturated_model, array_name), 0)


def test_pipeline_suffix(rtdata, ignore_asdf_paths):
    """
    Tests passing suffix to the pipeline

    Note that this test mimics how the pipeline is run in OPS.

    Any changes to this test should be coordinated with OPS.
    """
    input_data = "r0000101001001001001_0001_wfi01_f158_uncal.asdf"
    rtdata.get_data(f"WFI/image/{input_data}")

    output = "r0000101001001001001_0001_wfi01_f158_star.asdf"
    rtdata.output = output

    args = [
        "roman_elp",
        rtdata.input,
        "--steps.tweakreg.skip=True",
        "--suffix=star",
    ]
    ExposurePipeline.from_cmdline(args)
    rtdata.get_truth(f"truth/WFI/image/{output}")

    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()

    # Ensure step completion is as expected
    with rdm.open(rtdata.output) as model:
        assert model.meta.cal_step.dq_init == "COMPLETE"
        assert model.meta.cal_step.saturation == "COMPLETE"
        assert model.meta.cal_step.linearity == "COMPLETE"
        assert model.meta.cal_step.dark == "COMPLETE"
        assert model.meta.cal_step.ramp_fit == "COMPLETE"
        assert model.meta.cal_step.assign_wcs == "COMPLETE"
        assert model.meta.cal_step.flat_field == "COMPLETE"
        assert model.meta.cal_step.photom == "COMPLETE"
        assert model.meta.cal_step.source_catalog == "COMPLETE"
        assert model.meta.cal_step.tweakreg == "SKIPPED"
        assert model.meta.filename == output



"""Roman tests for source catalog creation"""

import pytest
from astropy.table import Table

from romancal.stpipe import RomanStep

# mark all tests in this module
pytestmark = [pytest.mark.bigdata, pytest.mark.soctests]

fieldlist = [
    "ra_centroid",  # DMS374 positions on ICRF
    "dec_centroid",  # DMS374 positions on ICRF
    "aper01_f158_flux",  # DMS399 aperture fluxes
    "aper02_f158_flux",  # DMS399 aperture fluxes
    "aper04_f158_flux",  # DMS399 aperture fluxes
    "segment_f158_flux",  # DMS375 fluxes
    "kron_f158_flux",  # DMS375 fluxes
    "aper01_f158_flux_err",  # DMS386 flux uncertainties
    "aper02_f158_flux_err",  # DMS386 flux uncertainties
    "warning_flags",  # DMS387 dq_flags
    "is_extended_f158",  # DMS392 source classification
    "semimajor",  # DMS394 galaxy morphology
    "semiminor",  # DMS394 galaxy morphology
    "orientation_sky",  # DMS394 galaxy morphology
    "segment_f158_flux_err",  # DMS395 basic statistical uncertainties
    "kron_f158_flux_err",  # DMS395 basic statistical uncertainties
    "aper01_f158_flux_err",  # DMS395 basic statistical uncertainties
    "x_psf_f158_err",  # DMS395 basic statistical uncertainties
]


def test_multiband_catalog(rtdata_module, resource_tracker, request, dms_logger):
    rtdata = rtdata_module
    inputasnfn = "L3_skycell_mbcat_asn.json"
    # note that this input association currently only has a single
    # filter in it, so this is more of an existence proof for the multiband
    # catalogs than a detailed test.  Using only a single catalog lets us
    # rely on the existing regtest files.
    outputfn = "r00001_p_v01001001001001_270p65x70y49_f158_mbcat_cat.parquet"
    rtdata.get_asn(f"WFI/image/{inputasnfn}")
    rtdata.output = outputfn
    rtdata.input = inputasnfn
    rtdata.get_truth(f"truth/WFI/image/{outputfn}")
    cattruth = Table.read(f"truth/{outputfn}")
    args = [
        "romancal.step.MultibandCatalogStep",
        inputasnfn,
        "--deblend",
        "True",  # use deblending, DMS 393
    ]
    with resource_tracker.track(log=request):
        RomanStep.from_cmdline(args)
    cat = Table.read(outputfn)
    for field in fieldlist:
        assert field in cat.dtype.names

    dms_logger.info(
        "DMS374, 399, 375, 386, 387, 392, 394, 395: source catalog includes fields: "
        + ", ".join(fieldlist)
    )

    # DMS 393: multiband catalog uses both PSF-like and extend-source-like
    # kernels
    assert set(cat.dtype.names) == set(cattruth.dtype.names)
    # weak assertion that our truth file must at least have the same
    # catalog fields as the file produced here.  Exactly matching rows
    # would require a lot of okifying things that aren't obviously
    # the same, but we can easily check that the columns match up
    # by name.

    dms_logger.info("DMS391: successfully used multiple kernels to detect sources.")
    dms_logger.info("DMS393: successfully used deblending to separate blended sources.")
    dms_logger.info(
        "DMS399: successfully tested that catalogs contain aperture "
        "fluxes and uncertainties."
    )



"""Access the Roman Engineering Mnemonic Database through MAST."""

import logging
from ast import literal_eval
from os import getenv

import numpy as np
import requests
from astropy.table import Table
from astropy.time import Time
from requests.adapters import HTTPAdapter, Retry

from .engdb_lib import (
    FORCE_STATUSES,
    RETRIES,
    TIMEOUT,
    EngdbABC,
    ValueCollection,
)

__all__ = ["EngdbMast"]

# Default MAST info.
MAST_BASE_URL = "https://masttest.stsci.edu"
API_URI = "edp/api/v0.1/mnemonics/fqa/roman/data"

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class EngdbMast(EngdbABC):
    """
    Access the Roman Engineering Database through MAST.

    Parameters
    ----------
    base_url : str
        The base url for the engineering RESTful service. If not defined,
        the environmental variable ENG_BASE_URL is queried. Otherwise
        the default MAST website is used.

    token : str or None
        The MAST access token. If not defined, the environmental variable
        MAST_API_TOKEN is queried. A token is required.
        For more information, see https://auth.mast.stsci.edu/

    **service_kwargs : dict
        Service-specific keyword arguments that are not relevant to this implementation
        of EngdbABC.

    Raises
    ------
    RuntimeError
        Any and all failures with connecting with the MAST server.
    """

    #: The base URL for the engineering service.
    base_url = None

    #: The end time of the last query.
    endtime = None

    #: The results of the last query.
    response = None

    #: Number of retries to attempt to contact the service
    retries = RETRIES

    #: The start time of the last query.
    starttime = None

    #: Network timeout when communicating with the service
    timeout = TIMEOUT

    #: MAST Token
    token = None

    def __init__(self, base_url=None, token=None, **service_kwargs):
        logger.debug("kwargs not used by this service: %s", service_kwargs)

        self.configure(base_url=base_url, token=token)

        # Check for basic aliveness.
        try:
            resp = requests.get(self.base_url + "edp/", timeout=self.timeout)
        except requests.exceptions.ConnectionError as exception:
            raise RuntimeError(
                f"MAST url: {self.base_url} is unreachable."
            ) from exception
        if resp.status_code != 200:
            raise RuntimeError(
                f"MAST url: {self.base_url} is not available. "
                f"Returned HTTPS status {resp.status_code}"
            )

        # Basics are covered. Finalize initialization.
        self.set_session()

    def configure(self, base_url=None, token=None):
        """
        Configure from parameters and environment.

        Parameters
        ----------
        base_url : str
            The base url for the engineering RESTful service. If not defined,
            the environmental variable ENG_BASE_URL is queried. Otherwise
            the default MAST website is used.

        token : str or None
            The MAST access token. If not defined, the environmental variable
            MAST_API_TOKEN is queried. A token is required.
            For more information, see 'https://auth.mast.stsci.edu/'
        """
        # Determine the database to use
        if base_url is None:
            base_url = getenv("ENG_BASE_URL", MAST_BASE_URL)
        if base_url[-1] != "/":
            base_url += "/"
        self.base_url = base_url

        # Get the token
        if token is None:
            token = getenv("MAST_API_TOKEN", None)
        self.token = token

        # Get various timeout parameters
        self.retries = getenv("ENG_RETRIES", RETRIES)
        self.timeout = getenv("ENG_TIMEOUT", TIMEOUT)

    def get_meta(self, *kwargs):
        """
        Get the mnemonics meta info.

        The MAST interface does not provide any meta.
        """
        raise NotImplementedError(
            "MAST Engineering AUI does not provide a meta service"
        )

    def get_values(
        self,
        mnemonic,
        starttime,
        endtime,
        time_format=None,
        include_obstime=False,
        include_bracket_values=False,
        zip_results=True,
    ):
        """
        Retrieve all results for a mnemonic in the requested time range.

        Parameters
        ----------
        mnemonic : str
            The engineering mnemonic to retrieve

        starttime : str or `astropy.time.Time`
            The, inclusive, start time to retrieve from.

        endtime : str or `astropy.time.Time`
            The, inclusive, end time to retrieve from.

        time_format : str
            The format of the input time used if the input times
            are strings. If None, a guess is made.

        include_obstime : bool
            If `True`, the return values will include observation
            time as `astropy.time.Time`. See `zip_results` for further details.

        include_bracket_values : bool
            The DB service, by default, returns the bracketing
            values outside of the requested time. If `True`, include
            these values.

        zip_results : bool
            If `True` and `include_obstime` is `True`, the return values
            will be a list of 2-tuples. If false, the return will
            be a single 2-tuple, where each element is a list.

        Returns
        -------
        values : [value, ...] or [(obstime, value), ...] or ([obstime,...], [value, ...])
            Returns the list of values. See `include_obstime` and `zip` for modifications.
        """
        if not isinstance(starttime, Time):
            starttime = Time(starttime, format=time_format)
        if not isinstance(endtime, Time):
            endtime = Time(endtime, format=time_format)

        records = self._get_records(
            mnemonic=mnemonic,
            starttime=starttime,
            endtime=endtime,
            time_format=time_format,
        )
        if len(records) == 0:
            return list()

        # If desired, remove bracket or outside of timeframe entries.
        if not include_bracket_values:
            selection = np.logical_and(
                records["MJD"] >= starttime.mjd, records["MJD"] <= endtime.mjd
            )
            records = records[selection]

        # Reformat to the desired list formatting.
        results = ValueCollection(
            include_obstime=include_obstime, zip_results=zip_results
        )
        values = records["EUValue"]
        obstimes = Time(records["MJD"], format="mjd")
        for obstime, value in zip(obstimes, values, strict=False):
            results.append(obstime, value)

        return results.collection

    def set_session(self):
        """Set up HTTP session."""
        self._req = requests.Request(
            method="GET",
            url=self.base_url + API_URI,
            headers={"Authorization": f"token {self.token}"},
        )

        s = requests.Session()
        retries = Retry(
            total=self.retries,
            backoff_factor=1.0,
            status_forcelist=FORCE_STATUSES,
            raise_on_status=True,
        )
        s.mount("https://", HTTPAdapter(max_retries=retries))

        self._session = s

    def _get_records(
        self, mnemonic, starttime, endtime, time_format=None, **other_kwargs
    ):
        """
        Retrieve all results for a mnemonic in the requested time range.

        Parameters
        ----------
        mnemonic : str
            The engineering mnemonic to retrieve

        starttime : str or astropy.time.Time
            The, inclusive, start time to retrieve from.

        endtime : str or astropy.time.Time
            The, inclusive, end time to retrieve from.

        time_format : str
            The format of the input time used if the input times
            are strings. If None, a guess is made.

        **other_kwargs : dict
            Keyword arguments not relevant to this implementation.

        Returns
        -------
        records : `astropy.Table`
            Returns the resulting table.

        Notes
        -----
        The engineering service always returns the bracketing entries
        before and after the requested time range.
        """
        if not isinstance(starttime, Time):
            starttime = Time(starttime, format=time_format)
        if not isinstance(endtime, Time):
            endtime = Time(endtime, format=time_format)
        self.starttime = starttime
        self.endtime = endtime

        # Make the request
        mnemonic = mnemonic.strip()
        mnemonic = mnemonic.upper()
        starttime_fmt = starttime.strftime("%Y-%m-%dT%H:%M:%S")
        endtime_fmt = endtime.strftime("%Y-%m-%dT%H:%M:%S")
        self._req.params = {
            "mnemonic": mnemonic,
            "s_time": starttime_fmt,
            "e_time": endtime_fmt,
        }
        prepped = self._session.prepare_request(self._req)
        settings = self._session.merge_environment_settings(
            prepped.url, {}, None, None, None
        )
        logger.debug("Query: %s", prepped.url)
        self.response = self._session.send(prepped, timeout=self.timeout, **settings)
        self.response.raise_for_status()
        logger.debug("Response: %s", self.response)
        logger.debug("Response test: %s", self.response.text)

        # Convert to table.
        response = literal_eval(self.response.text)
        data = response["Data"]
        del response["Data"]
        table = Table(rows=data, meta=response)

        # Create a column MJD that has the MJD version of the data
        if len(table):
            obstime = Time(table["ObsTime"])
            table["MJD"] = obstime.mjd

        return table

    def __repr__(self):
        """What am I"""
        repr = f"{self.__class__.__name__}(base_url='{self.base_url}')"
        return repr



#!/usr/bin/env python
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import romancal.datamodels.filetype as filetype
from romancal.datamodels import ModelLibrary

# step imports
from romancal.flux import FluxStep
from romancal.outlier_detection import OutlierDetectionStep
from romancal.resample import ResampleStep
from romancal.skymatch import SkyMatchStep
from romancal.source_catalog import SourceCatalogStep

from ..stpipe import RomanPipeline

if TYPE_CHECKING:
    from typing import ClassVar

__all__ = ["MosaicPipeline"]

# Define logging
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class MosaicPipeline(RomanPipeline):
    """
    MosaicPipeline: Apply all calibration steps to the roman data
    to produce level 3 products. Included steps are:
    ``flux``, ``skymatch``, ``outlier_detection``, ``resample`` and ``source catalog``.
    """

    class_alias = "roman_mos"
    spec = """
        save_results = boolean(default=False)
        on_disk = boolean(default=False)
        resample_on_skycell = boolean(default=True)
    """

    # Define aliases to steps
    step_defs: ClassVar = {
        "flux": FluxStep,
        "skymatch": SkyMatchStep,
        "outlier_detection": OutlierDetectionStep,
        "resample": ResampleStep,
        "source_catalog": SourceCatalogStep,
    }

    # start the actual processing
    def process(self, input_data):
        """Process the Roman WFI data from Level 2 to Level 3"""

        log.info("Starting Roman mosaic level calibration pipeline ...")

        # open the input file
        file_type = filetype.check(input_data)
        if file_type == "asn":
            library = ModelLibrary(input_data, on_disk=self.on_disk)
        elif file_type == "ModelLibrary":
            library = input_data
        else:
            raise TypeError(
                "The level three pipeline input needs to be an association or ModelLibrary"
            )

        # propagate resample_on_skycell setting
        self.outlier_detection.resample_on_skycell = self.resample_on_skycell
        self.resample.resample_on_skycell = self.resample_on_skycell

        self.flux.suffix = "flux"
        result = self.flux.run(library)
        self.skymatch.suffix = "skymatch"
        result = self.skymatch.run(result)
        self.outlier_detection.suffix = "outlier_detection"
        result = self.outlier_detection.run(result)
        self.resample.suffix = "coadd"
        self.output_file = library.asn["products"][0]["name"]
        result = self.resample.run(result)
        self.source_catalog.output_file = self.output_file
        self.source_catalog.run(result)
        self.suffix = "coadd"
        return result



"""Test module pointing_summary"""

import sys
from pathlib import Path

import pytest
from astropy.table import Table
from astropy.time import Time
from astropy.utils.diff import report_diff_values
from roman_datamodels.datamodels import ScienceRawModel

import romancal.orientation.pointing_summary as ps
from romancal.lib.engdb import engdb_tools

DATA_PATH = Path(__file__).parent / "data"

# Engineering parameters
GOOD_STARTTIME = Time("2027-02-23T01:00:00", format="isot")
GOOD_ENDTIME = Time("2027-02-23T01:00:05", format="isot")

# Header defaults
TARG_RA = 270.0
TARG_DEC = 66.01


@pytest.fixture
def engdb():
    """Setup the service to operate through the mock service"""
    try:
        engdb = engdb_tools.engdb_service(service="mast")
    except RuntimeError as exception:
        pytest.skip(f"Engineering database unvailable: {exception}.")
    yield engdb


@pytest.fixture(scope="module")
def model_path(module_jail):
    """Create data file with needed header parameters"""
    model = ScienceRawModel.create_fake_data(
        {
            "meta": {
                "exposure": {"start_time": GOOD_STARTTIME, "end_time": GOOD_ENDTIME},
                "pointing": {"target_ra": TARG_RA, "target_dec": TARG_DEC},
            }
        }
    )

    model.meta.pointing.target_ra = TARG_RA
    model.meta.pointing.target_dec = TARG_DEC
    model.meta.pointing.ra_v1 = 91.08142005
    model.meta.pointing.dec_v1 = -66.60547869
    model.meta.wcsinfo.ra_ref = 90.70377653
    model.meta.wcsinfo.dec_ref = -66.59540224

    model_path = module_jail / "model.asdf"
    model.save(model_path)
    return model_path


def test_calc_pointing_deltas(engdb, model_path):
    """Test `calc_pointing_deltas` basic running"""
    truth = (
        "Delta(target=<SkyCoord (ICRS): (ra, dec) in deg"
        "\n    (270., 66.01)>, v1=<SkyCoord (ICRS): (ra, dec) in deg"
        "\n    (91.08142005, -66.60547869)>, refpoint=<SkyCoord (ICRS): (ra, dec) in deg"
        "\n    (90.70377653, -66.59540224)>, delta_v1=<Angle 179.26285175 deg>, delta_refpoint=<Angle 179.34985532 deg>)"
    )
    with ScienceRawModel(str(model_path)) as model:
        deltas = ps.calc_pointing_deltas(model)

    assert truth == str(deltas)


def test_calc_deltas(engdb, model_path, tmp_path):
    """Test `calc_deltas` basic running"""
    with ScienceRawModel(model_path) as model:
        deltas = ps.calc_deltas([model])

    # Save the calculated deltas for test debugging
    deltas["exposure"] = [m.meta.filename for m in deltas["exposure"]]
    deltas.write(tmp_path / "calc_deltas_truth.ecsv", format="ascii.ecsv")

    truth = Table.read(DATA_PATH / "calc_deltas_truth.ecsv")

    # round the delta values to a reasonable level
    deltas[0][4] = round(deltas[0][4], 8)
    deltas[0][5] = round(deltas[0][5], 8)
    truth[0][4] = round(truth[0][4], 8)
    truth[0][5] = round(truth[0][5], 8)

    assert report_diff_values(truth, deltas, fileobj=sys.stderr)



import pytest
import roman_datamodels as rdm

from romancal.pipeline.mosaic_pipeline import MosaicPipeline
from romancal.resample.l3_wcs import l3wcsinfo_to_wcs

from . import util
from .regtestdata import compare_asdf

# mark all tests in this module
pytestmark = [pytest.mark.bigdata, pytest.mark.soctests]


@pytest.fixture(scope="module")
def run_mos(rtdata_module, resource_tracker):
    rtdata = rtdata_module

    # Test Pipeline
    rtdata.get_asn("WFI/image/L3_mosaic_asn.json")
    output = "r00001_p_v01001001001001_270p65x70y49_f158_coadd.asdf"
    rtdata.output = output
    args = [
        "roman_mos",
        rtdata.input,
    ]
    with resource_tracker.track():
        MosaicPipeline.from_cmdline(args)
    rtdata.get_truth(f"truth/WFI/image/{output}")
    return rtdata


@pytest.fixture(scope="module")
def output_filename(run_mos):
    return run_mos.output


@pytest.fixture(scope="module")
def output_model(output_filename):
    with rdm.open(output_filename) as model:
        yield model


@pytest.fixture(scope="module")
def truth_filename(run_mos):
    return run_mos.truth


def test_log_tracked_resources(log_tracked_resources, run_mos):
    log_tracked_resources()


def test_output_matches_truth(output_filename, truth_filename, ignore_asdf_paths):
    diff = compare_asdf(output_filename, truth_filename, **ignore_asdf_paths)
    assert diff.identical, diff.report()


def test_resample_ran(output_model):
    # DMS373 test output is resampled to a skycell
    # FIXME this doesn't test if the output is a skycell
    assert output_model.meta.cal_step.resample == "COMPLETE"


def test_location_name(output_model):
    # test that the location_name matches the skycell selected
    assert output_model.meta.wcsinfo.skycell_name == "270p65x70y49"


def test_wcsinfo_wcs_roundtrip(output_model):
    """Test that the contents of wcsinfo reproduces the wcs"""
    gwcs_from_wcsinfo = l3wcsinfo_to_wcs(output_model.meta.wcsinfo)

    ra_mad, dec_mad = util.comp_wcs_grids_arcs(output_model.meta.wcs, gwcs_from_wcsinfo)
    assert (ra_mad + dec_mad) / 2.0 < 1.0e-4



import numpy as np


def copy_mosaic_meta(model, cat_model):
    # TODO some junk values here
    cat_model.meta.prd_version = "8.8.8"
    cat_model.meta.sdf_software_version = "7.7.7"
    cat_model.meta.basic.time_first_mjd = model.meta.coadd_info.time_first.mjd
    cat_model.meta.basic.time_last_mjd = model.meta.coadd_info.time_last.mjd
    cat_model.meta.basic.time_mean_mjd = model.meta.coadd_info.time_mean.mjd
    cat_model.meta.basic.max_exposure_time = model.meta.coadd_info.get(
        "max_exposure_time", np.nan
    )
    cat_model.meta.basic.mean_exposure_time = model.meta.coadd_info.exposure_time
    cat_model.meta.basic.visit = model.meta.observation.visit
    cat_model.meta.basic.segment = model.meta.observation.segment
    cat_model.meta.basic["pass"] = model.meta.observation["pass"]
    cat_model.meta.basic.program = model.meta.observation.program
    cat_model.meta.basic.survey = "?"

    # TODO handle optical element with multiple values
    cat_model.meta.basic.optical_element = model.meta.instrument.optical_element

    cat_model.meta.basic.instrument = "WFI"
    cat_model.meta.basic.location_name = model.meta.wcsinfo.skycell_name
    cat_model.meta.basic.product_type = model.meta.product_type

    # TODO can we fill these in?
    cat_model.meta.photometry.conversion_megajanskys = None
    cat_model.meta.photometry.conversion_megajanskys_uncertainty = None
    cat_model.meta.photometry.pixel_area = None
    return cat_model



import logging
import os
import re
from datetime import datetime
from functools import cached_property
from pathlib import Path

import crds
import numpy as np
import roman_datamodels
import spherical_geometry.great_circle_arc as sga
import spherical_geometry.polygon as sgp
import spherical_geometry.vector as sgv
from asdf import AsdfFile
from gwcs import WCS
from numpy.typing import NDArray
from scipy.spatial import KDTree

from romancal.datamodels.library import ModelLibrary
from romancal.lib.wcsinfo_to_wcs import wcsinfo_to_wcs

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


__all__ = ["SKYMAP", "ProjectionRegion", "SkyCell", "SkyMap"]


class SkyCell:
    """Square subregion of a projection region, 4.6 arcminutes per side."""

    _index: int | None
    _data: np.void
    _skymap: "SkyMap"

    # average area of a sky cell in square degrees on the sphere
    area = 1.7760288493318122e-06

    # average diagonal length of a skycell in degrees on the sphere
    length = 0.0018846729895155984

    def __init__(self, index: int | None, skymap: "SkyMap" = None):
        """
        Parameters
        ----------
        index : int
            Index in the global sky map.
        skymap: SkyMap
            sky map instance (defaults to global SKYMAP)
        """

        if skymap is None:
            skymap = SKYMAP

        self._index = index
        self._skymap = skymap
        self._data = None

    @classmethod
    def from_name(cls, name: str, skymap: "SkyMap" = None) -> "SkyCell":
        """Retrieve a sky cell from the sky map by its name (see handbook [1] for explanation).

        Parameters
        ----------
        name : str
            Name of a sky cell, for instance `315p86x50y75`.
        skymap : SkyMap
            sky map instance; defaults to global SKYMAP (Default value = None)

        References
        ----------
        .. [1] `Skymap Tessellation <https://roman-docs.stsci.edu/data-handbook-home/wfi-data-format/skymap-tessellation>`_
        """
        if not re.match(r"\d{3}\w\d{2}x\d{2}y\d{2}", name):
            raise ValueError(f"invalid skycell name {name}")

        if skymap is None:
            skymap = SKYMAP

        indices = np.where(skymap.model.skycells["name"] == name)[0]
        if len(indices) > 0:
            return SkyCell(indices[0], skymap=skymap)
        else:
            raise KeyError(
                f"skycell with name '{name}' does not exist in the currently-loaded sky map"
            )

    @classmethod
    def from_data(cls, data: np.void, skymap: "SkyMap" = None) -> "SkyCell":
        """build an index-less sky cell instance from a data array

        Parameters
        ----------
        data : np.void
            array with sky cell parameters (see schema)
        skymap: SkyMap
            sky map instancel; defaults to global SKYMAP (Default value = None)
        """
        instance = cls(index=None, skymap=skymap)
        instance._data = data
        return instance

    @classmethod
    def from_asn(cls, asn: dict | str, skymap: "SkyMap" = None) -> "SkyCell":
        """retrieve a sky cell from WCS info or a target specified in an association

        Attempts to find a sky cell name from the following in order:
            - `skycell_wcs_info.name`
            - `target`

        Parameters
        ----------
        asn : dict | str
            association dictionary or a path to an association file to load
        skymap: SkyMap
            sky map instance; defaults to global SKYMAP (Default value = None)
        """

        if isinstance(asn, str | os.PathLike):
            asn = ModelLibrary._load_asn(asn)

        skycell_name = None
        if "skycell_wcs_info" in asn and isinstance(asn["skycell_wcs_info"], dict):
            skycell_name = asn["skycell_wcs_info"]["name"]
        elif "target" in asn and asn["target"].lower() != "none":
            skycell_name = asn["target"]
        else:
            raise ValueError(
                "cannot extract skycell information from an association without `skycell_wcs_info` or `target`"
            )

        return SkyCell.from_name(skycell_name, skymap=skymap)

    @property
    def index(self) -> int | None:
        """index of this sky cell in the sky map"""
        return self._index

    @property
    def data(self) -> np.void:
        """Sky cell data.

        ("name", "ra_center", "dec_center", "orientat", "x_tangent", "y_tangent", "ra_corn1", "dec_corn1", "ra_corn2", "dec_corn2", "ra_corn3", "dec_corn3", "ra_corn4", "dec_corn4")
        """
        if self._data is None and self.index is not None:
            self._data = self._skymap.model.skycells[self.index]
        return self._data

    @property
    def name(self) -> str:
        """name of this sky cell, for instance `315p86x50y75`

        NOTE
        ----
        the name of a sky cell center comprises the rounded center coordinates of its containing projection region in right ascension and declination,
        and the XY location of the sky cell within its projection region in units of ordinal sky cells from that center
        """
        return self.data["name"].item()

    @property
    def radec_center(self) -> tuple[float, float]:
        """center point in right ascension and declination"""
        return self.data["ra_center"].item(), self.data["dec_center"].item()

    @property
    def orientation(self) -> float:
        return self.data["orientat"].item()

    @property
    def xy_tangent(self) -> tuple[float, float]:
        """center point in pixel coordinates"""
        return self.data["x_tangent"].item(), self.data["y_tangent"].item()

    @property
    def radec_corners(
        self,
    ) -> NDArray[float]:
        """corners in right ascension and declination in the order given by the sky map"""
        return np.array(
            (
                self.data[["ra_corn1", "dec_corn1"]].item(),
                self.data[["ra_corn2", "dec_corn2"]].item(),
                self.data[["ra_corn3", "dec_corn3"]].item(),
                self.data[["ra_corn4", "dec_corn4"]].item(),
            )
        )

    @cached_property
    def vectorpoint_corners(self) -> NDArray[float]:
        """corners in 3D Cartesian space on the unit sphere"""
        return sgv.normalize_vector(
            np.stack(sgv.lonlat_to_vector(*np.array(self.radec_corners).T), axis=1)
        )

    @cached_property
    def vectorpoint_center(self) -> tuple[float, float, float]:
        """center in 3D Cartesian space on the unit sphere"""
        return sgv.normalize_vector(np.array(sgv.lonlat_to_vector(*self.radec_center)))

    @cached_property
    def polygon(self) -> sgp.SingleSphericalPolygon:
        """spherical polygon representing this sky cell"""
        return sgp.SingleSphericalPolygon(
            points=self.vectorpoint_corners,
            inside=self.vectorpoint_center,
        )

    @cached_property
    def projection_region(self) -> "ProjectionRegion":
        """Projection region containing this sky cell."""
        if self.index is None:
            raise ValueError("no index provided")
        return ProjectionRegion.from_skycell_index(self.index)

    @property
    def pixel_scale(self) -> float:
        """degrees per pixel"""
        return self._skymap.pixel_scale

    @property
    def pixel_shape(self) -> tuple[int, int]:
        """number of pixels across"""
        return self._skymap.pixel_shape

    @property
    def wcs_info(self) -> dict[str, float | str]:
        """WCS properties as defined in the Level 3 association schema"""

        return {
            "name": self.name,
            "pixel_scale": self.pixel_scale,
            "ra_projection_center": self.projection_region.radec_tangent[0],
            "dec_projection_center": self.projection_region.radec_tangent[1],
            "x0_projection": self.xy_tangent[0],
            "y0_projection": self.xy_tangent[1],
            "ra_center": self.radec_center[0],
            "dec_center": self.radec_center[1],
            "nx": self.pixel_shape[0],
            "ny": self.pixel_shape[1],
            "orientat": self.orientation,
            "orientat_projection_center": self.projection_region.orientation,
        }

    @cached_property
    def wcs(self) -> WCS:
        """WCS representing this sky cell"""
        wcsobj = wcsinfo_to_wcs(
            self.wcs_info,
            bounding_box=(
                (-0.5, self.pixel_shape[0] - 0.5),
                (-0.5, self.pixel_shape[1] - 0.5),
            ),
        )
        wcsobj.array_shape = self.pixel_shape
        return wcsobj

    def __eq__(self, other) -> bool:
        if not isinstance(other, SkyCell):
            return False

        return self.data == other.data

    def __str__(self) -> str:
        return f"{self.name} [{self._skymap.path}]"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.index}, {self._skymap})"


class ProjectionRegion:
    """Projection region in the sky map."""

    _index: int | None
    _data: np.void
    _skymap: "SkyMap"

    # area of the smallest projection region in square degrees on the sphere
    #   min(sc.ProjectionRegion(index).polygon.area() for index in range(len(sc.SKYMAP.model.projection_regions)))
    MIN_AREA = 0.002791388883915502

    # diagonal length of the longest projection region in degrees on the sphere
    #   max(sc.ProjectionRegion(index).length for index in range(len(sc.SKYMAP.model.projection_regions)))
    MAX_LENGTH = 0.08174916691321586

    def __init__(self, index: int | None, skymap: "SkyMap" = None):
        """
        Parameters
        ----------
        index : int
            index of the projection region in the sky map array
        skymap: SkyMap
            sky map instance (defaults to global SKYMAP)
        """

        if skymap is None:
            skymap = SKYMAP

        self._index = index
        self._skymap = skymap
        if index is not None:
            self._data = self._skymap.model.projection_regions[index]

    @classmethod
    def from_data(cls, data: np.void, skymap: "SkyMap" = None) -> "ProjectionRegion":
        """build a projection region instance from a data array

        Parameters
        ----------
        data : numpy.void
            array with projection region parameters (see schema)
        skymap: SkyMap
            sky map instance; defaults to global SKYMAP (Default value = None)
        """
        instance = cls(index=data["index"], skymap=skymap)
        instance._data = data
        return instance

    @classmethod
    def from_skycell_index(
        cls, index: int, skymap: "SkyMap" = None
    ) -> "ProjectionRegion":
        """
        Parameters
        ----------
        index : int
            index of the sky cell
        skymap : SkyMap
            sky map instance; defaults to global SKYMAP (Default value = None)

        Returns
        -------
        ProjectionRegion
            projection region corresponding to the given sky cell
        """

        if skymap is None:
            skymap = SKYMAP

        projregion_indices = np.where(
            (index >= skymap.model.projection_regions["skycell_start"])
            & (index < skymap.model.projection_regions["skycell_end"])
        )[0]
        if len(projregion_indices) == 1:
            return cls(projregion_indices[0], skymap=skymap)
        else:
            msg = (
                f"sky cell index {index} not found in any projection regions"
                if len(projregion_indices) == 0
                else f"sky cell index {index} found in multiple projection regions; possibly malformed sky map at {skymap.path}"
            )
            raise KeyError(msg)

    @property
    def index(self) -> int | None:
        """index in the sky map"""
        return self._index

    @property
    def data(self) -> np.void:
        """Projection region data.

        ("index", "ra_tangent", "dec_tangent", "ra_min", "ra_max", "dec_min", "dec_max", "orientat", "x_tangent", "y_tangent", "nx", "ny", "skycell_start", "skycell_end")
        """
        return self._data

    @property
    def radec_tangent(self) -> tuple[float, float]:
        """projection origin (tangent point with the celestial sphere) in right ascension and declination"""
        return self.data[["ra_tangent", "dec_tangent"]].item()

    @property
    def radec_bounds(self) -> tuple[float, float, float, float]:
        """bounds in right ascension and declination in order [xmin, ymin, xmax, ymax]"""
        return self.data[["ra_min", "dec_min", "ra_max", "dec_max"]].item()

    @property
    def orientation(self) -> float:
        return self.data["orientat"].item()

    @property
    def xy_tangent(self) -> tuple[float, float]:
        """projection origin (tangent point with the celestial sphere) in pixel coordinates"""
        return self.data[["x_tangent", "y_tangent"]].item()

    @property
    def pixel_shape(self) -> tuple[int, int]:
        """number of pixels across"""
        return self.data[["nx", "ny"]].item()

    @cached_property
    def skycell_indices(self) -> NDArray[int]:
        """indices of sky cells in the sky map within this region"""
        return np.arange(self.data["skycell_start"], self.data["skycell_end"])

    @property
    def skycells(self) -> np.void:
        """subset array of sky cells from the sky map within this region"""
        return self._skymap.model.skycells[self.skycell_indices]

    @cached_property
    def skycells_kdtree(self) -> KDTree:
        """LOCAL k-d tree of skycells in this projection region, using normalized center vectorpoints in 3D space

        NOTE
        ----
        add `skycell_start` to the indices returned by this tree to convert to skycell indices in the parent skymap
        """

        return KDTree(
            sgv.normalize_vector(
                np.stack(
                    sgv.lonlat_to_vector(
                        self.skycells["ra_center"],
                        self.skycells["dec_center"],
                    ),
                    axis=1,
                )
            )
        )

    @property
    def radec_corners(
        self,
    ) -> NDArray:
        """corners in right ascension and declination in clockwise order"""
        return np.array(
            (
                self.data[["ra_min", "dec_min"]].item(),
                self.data[["ra_max", "dec_min"]].item(),
                self.data[["ra_max", "dec_max"]].item(),
                self.data[["ra_min", "dec_max"]].item(),
            )
        )

    @cached_property
    def vectorpoint_corners(self) -> NDArray[float]:
        """corners in 3D Cartesian space on the unit sphere"""
        return sgv.normalize_vector(
            np.stack(sgv.lonlat_to_vector(*np.array(self.radec_corners).T), axis=1)
        )

    @cached_property
    def vectorpoint_center(self) -> tuple[float, float, float]:
        """center in 3D Cartesian space on the unit sphere"""
        return np.mean(self.vectorpoint_corners, axis=0)

    @cached_property
    def length(self) -> float:
        """diagonal length of the region"""
        # assume radial against sky background
        return max(
            sga.length(
                self.vectorpoint_corners[index], self.vectorpoint_corners[index + 2]
            )
            for index in range(len(self.vectorpoint_corners) - 3)
        )

    @property
    def is_polar(self) -> bool:
        """whether this projection region is a polar cap"""
        return self.data["dec_max"] == 90.0 or self.data["dec_min"] == -90.0

    @cached_property
    def polygon(self) -> sgp.SingleSphericalPolygon:
        """spherical polygon representing this region"""
        if self.is_polar:
            # the projection regions at the poles are circular caps on the sphere;
            # a polygon built from the corners in that case would be degenerate
            return sgp.SingleSphericalPolygon.from_cone(
                *self.radec_tangent,
                radius=(
                    90.0 - self.radec_bounds[1]
                    if self.data["dec_max"] == 90.0
                    else -self.radec_bounds[3]
                ),
                steps=16,
            )
        else:
            return sgp.SingleSphericalPolygon(
                points=self.vectorpoint_corners,
                inside=self.vectorpoint_center,
            )

    @property
    def pixel_scale(self) -> float:
        """degrees per pixel"""
        return self._skymap.pixel_scale

    def __eq__(self, other) -> bool:
        if not isinstance(other, ProjectionRegion):
            return False

        return self.data == other.data

    def __str__(self) -> str:
        return f"projregion {self.index} [{self._skymap.path}]"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.index}, {self._skymap})"


class SkyMap:
    """Abstract representation of the sky map, comprising of 4058 overlapping rectangular
    "projection regions" defining gnomonic projection on to uniform pixel grids.
    The pixel scale for all projection regions is identical.

    Each projection region is subdivided into ~2000 square subregions ("sky cells", ~8 million in total),
    each 4.6' across. These sky cells also overlap each other by a standard number of pixels.

    References
    ----------
    .. [1] `Skymap Tessellation <https://roman-docs.stsci.edu/data-handbook-home/wfi-data-format/skymap-tessellation>`_
    """

    _path: None | Path
    _data: AsdfFile

    def __init__(self, path: None | Path | str = None):
        """
        Parameters
        ----------
        path : None | Path | str, optional
            load sky map from the specified ASDF file (defaults to latest `skycells` ref on CRDS)
        """
        if path is not None and not isinstance(path, Path):
            path = Path(path)
        self._path = path
        self._data = None

    @property
    def path(self) -> None | Path:
        """location of sky map reference file on filesystem"""
        return self._path

    @path.setter
    def path(self, path: None | Path):
        self._path = path
        # reset data if retrieved
        self._data = None

    @property
    def model(self) -> AsdfFile:
        """data model of sky map"""
        if self._data is None:
            if self._path is None:
                rmap = crds.getreferences(
                    {
                        "roman.meta.instrument.name": "WFI",
                        "roman.meta.exposure.start_time": f"{datetime.now():%Y-%m-%d %H:%M:%S}",
                    },
                    reftypes=["skycells"],
                    observatory="roman",
                )
                self._path = Path(rmap["skycells"])
            self._data = roman_datamodels.open(self._path, memmap=True)
        return self._data

    @cached_property
    def skycells_kdtree(self) -> KDTree:
        """k-d tree of all skycells in the skymap, using normalized center vectorpoints in 3D space

        NOTE
        ----
        there are 8 million skycells in the skymap; constructing this tree will take a long time. It is recommended that you instead use the `.skycells_kdtree` property of an individual projection region instead.
        """
        return KDTree(
            sgv.normalize_vector(
                np.stack(
                    sgv.lonlat_to_vector(
                        self.model.skycells["ra_center"],
                        self.model.skycells["dec_center"],
                    ),
                    axis=1,
                )
            )
        )

    @cached_property
    def projection_regions_kdtree(self) -> KDTree:
        """k-d tree of all projection regions in the skymap, using normalized center vectorpoints in 3D space"""
        return KDTree(
            sgv.normalize_vector(
                np.stack(
                    sgv.lonlat_to_vector(
                        self.model.projection_regions["ra_tangent"],
                        self.model.projection_regions["dec_tangent"],
                    ),
                    axis=1,
                )
            )
        )

    @property
    def pixel_scale(self) -> float:
        """degrees per pixel"""
        # The scale is given in arcseconds per pixel. Convert to degrees.
        return self.model.meta["pixel_scale"] / 3600.0

    @property
    def pixel_shape(self) -> tuple[int, int]:
        """number of pixels per sky cell"""
        return self.model.meta.nxy_skycell, self.model.meta.nxy_skycell

    def __getitem__(self, index: int) -> SkyCell:
        """`SkyCell` at the given index in the sky cells array"""
        return SkyCell(index)

    def __str__(self) -> str:
        return f"skymap {self.path}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.path})"


SKYMAP = SkyMap(path=os.environ.get("SKYMAP_PATH", None))



"""Tests for the Roman PSF Library"""

import numpy as np
import pytest
from roman_datamodels import datamodels as rdm

from romancal.source_catalog import psf
from romancal.step import SourceCatalogStep

# mark all tests in this module
pytestmark = [pytest.mark.bigdata, pytest.mark.soctests]


@pytest.fixture(
    scope="module",
    params=[
        "r0000101001001001001_0001_wfi01_f158_cal.asdf",
    ],
    ids=["input_file"],
)
def get_input_file(rtdata_module, request, resource_tracker):
    """Get the input file for testing"""

    rtdata = rtdata_module

    input_file = request.param

    rtdata.get_data(f"WFI/image/{input_file}")
    rtdata.input = input_file
    return rtdata


@pytest.mark.bigdata
def test_psf_library_reffile(get_input_file, dms_logger):
    """Test the retrieval of the PSF reference file"""

    rtdata = get_input_file

    # DMS 531 is to check that a PSF library has been provided and
    # DMS 532 is that we have access to the PSF library. By retrieving a
    # PSF reference file we show that the library exists and we have access.
    step = SourceCatalogStep()
    ref_file = step.get_reference_file(rtdata.input, "epsf")
    has_ref_file = ref_file != "N/A"
    assert has_ref_file
    passmsg = "PASS" if has_ref_file else "FAIL"
    dms_logger.info(f"DMS531 {passmsg},  ePSF reference file {ref_file} exists.")
    dms_logger.info(f"DMS532 {passmsg},  ePSF reference file {ref_file} was retrieved.")


@pytest.mark.bigdata
def test_psf_library_crdsfile(get_input_file, dms_logger):
    """Test that the PSF reference file matches the observation"""
    # DMS 535 identify an appropriate PSF for WFI source
    # check that the detector and optical element for the psf ref file
    # matches the data file

    rtdata = get_input_file
    input_data = rtdata.input

    step = SourceCatalogStep()
    ref_file = step.get_reference_file(input_data, "epsf")
    with rdm.open(ref_file) as ref_data:
        with rdm.open(rtdata.input) as wfi_data:
            psf_selection_match = (
                wfi_data.meta.instrument.optical_element
                == ref_data.meta.instrument.optical_element
            ) and (
                wfi_data.meta.instrument.detector == ref_data.meta.instrument.detector
            )

            assert psf_selection_match
            passmsg = "PASS" if psf_selection_match else "FAIL"
            dms_logger.info(
                f"DMS535 {passmsg},  ePSF reference file selection matches input data file"
            )


@pytest.mark.bigdata
def test_psf_library_psfinterp(get_input_file, dms_logger):
    """Test that the interpolation is occurring"""
    # DMS 536 interpolating empirical ePSFs given the observed-source position
    # Create two psf's one at the center and one off-center to show we can interpolate the PSF
    # and that the two are not the same

    rtdata = get_input_file
    input_data = rtdata.input

    step = SourceCatalogStep()
    ref_file = step.get_reference_file(input_data, "epsf")
    with rdm.open(ref_file) as ref_data:
        psf_model = psf.get_gridded_psf_model(ref_data)
        center = 2044
        szo2 = 10  # psf stamp axis lenght/2
        oversample = 11
        npix = szo2 * 2 * oversample + 1
        pts = np.linspace(-szo2 + center, szo2 + center, npix)
        xx, yy = np.meshgrid(pts, pts)
        psf_center = psf_model.evaluate(xx, yy, 1, center, center)
        off_center = center + 99
        pts = np.linspace(-szo2 + off_center, szo2 + off_center, npix)
        xx, yy = np.meshgrid(pts, pts)
        psf_offcenter = psf_model.evaluate(xx, yy, 1, off_center, off_center)
        # show that these two are different
        diff = psf_offcenter - psf_center
        # check to make sure the difference is not zero over the array
        psf_diff = diff.any()
        assert psf_diff
        passmsg = "PASS" if psf_diff else "FAIL"
        dms_logger.info(
            f"DMS536 {passmsg},  The interpolated ePSF for two positions are not the same"
        )



"""Access the Roman Engineering Mnemonic Database through the direct EDP interface."""

import logging
from os import getenv

import numpy as np
import requests
from astropy.table import Table
from astropy.time import Time
from requests.adapters import HTTPAdapter, Retry

from .engdb_lib import (
    FORCE_STATUSES,
    RETRIES,
    TIMEOUT,
    EngdbABC,
    ValueCollection,
)

__all__ = ["EngdbEDP"]

# Default MAST info.
MAST_BASE_URL = "https://masttest.stsci.edu"
API_URI = "edp/api/v0.1/mnemonics/fqa/roman/data"

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class EngdbEDP(EngdbABC):
    """
    Access the Roman Engineering Database through the direct EDP interface.

    The EDP interface is maintained by STScI MAST.

    Two items are required. An `edp.keytab` for authentication to the EDP servers.

    Environment variable MSSQL_DRIVER needs to point to the Microsoft SQL driver library.

    Parameters
    ----------
    environment : ['dev', 'test', 'int', 'ops']
        The operating environment currently working under.

    path_to_cc : Path-like
        Path to the kerberos keytab required to access the database.

    **service_kwargs : dict
        Service-specific keyword arguments that are not relevant to this implementation
        of EngdbABC.

    Raises
    ------
    RuntimeError
        Any and all failures with connecting with the MAST server.
    """

    #: The end time of the last query.
    endtime = None

    #: The MnemonicsReader instance
    mr = None

    #: Operating environment
    ops_env = None

    #: Path to the keytab
    path_to_cc: None

    #: The results of the last query.
    response = None

    #: Number of retries to attempt to contact the service
    retries = RETRIES

    #: The start time of the last query.
    starttime = None

    #: Network timeout when communicating with the service
    timeout = TIMEOUT

    def __init__(self, environment, path_to_cc, **service_kwargs):
        logger.debug("kwargs not used by this service: %s", service_kwargs)

        try:
            self.configure(environment, path_to_cc)
        except (FileNotFoundError, ModuleNotFoundError) as exception:
            raise RuntimeError(
                f"Cannot instantiate EDP service with environment: {environment}, path_to_cc: {path_to_cc}"
            ) from exception

        # Check for basic aliveness.
        self.mr.get_mnemonic_metadata()

    def configure(self, ops_env, path_to_cc):
        """
        Configure from parameters and environment.

        Parameters
        ----------
        environment : ['dev', 'test', 'int', 'ops']
            The operating environment currently working under.

        path_to_cc : Path-like
            Path to the kerberos keytab required to access the database.
        """
        self.ops_env = ops_env
        self.path_to_cc = path_to_cc

        # Create the mnemonic reader
        from edp.mnemonics_reader import MnemonicsReader

        self.mr = MnemonicsReader(
            mission="roman", env=ops_env, access_type="FqA", path_to_cc=path_to_cc
        )

        # Get various timeout parameters
        self.retries = getenv("ENG_RETRIES", RETRIES)
        self.timeout = getenv("ENG_TIMEOUT", TIMEOUT)

    def get_meta(self, *kwargs):
        """
        Get the mnemonics meta info.

        The MAST interface does not provide any meta.
        """
        raise NotImplementedError(
            "MAST Engineering AUI does not provide a meta service"
        )

    def get_values(
        self,
        mnemonic,
        starttime,
        endtime,
        time_format=None,
        include_obstime=False,
        include_bracket_values=False,
        zip_results=True,
    ):
        """
        Retrieve all results for a mnemonic in the requested time range.

        Parameters
        ----------
        mnemonic : str
            The engineering mnemonic to retrieve

        starttime : str or `astropy.time.Time`
            The, inclusive, start time to retrieve from.

        endtime : str or `astropy.time.Time`
            The, inclusive, end time to retrieve from.

        time_format : str
            The format of the input time used if the input times
            are strings. If None, a guess is made.

        include_obstime : bool
            If `True`, the return values will include observation
            time as `astropy.time.Time`. See `zip_results` for further details.

        include_bracket_values : bool
            The DB service, by default, returns the bracketing
            values outside of the requested time. If `True`, include
            these values.

        zip_results : bool
            If `True` and `include_obstime` is `True`, the return values
            will be a list of 2-tuples. If false, the return will
            be a single 2-tuple, where each element is a list.

        Returns
        -------
        values : [value, ...] or [(obstime, value), ...] or ([obstime,...], [value, ...])
            Returns the list of values. See `include_obstime` and `zip` for modifications.
        """
        if not isinstance(starttime, Time):
            starttime = Time(starttime, format=time_format)
        if not isinstance(endtime, Time):
            endtime = Time(endtime, format=time_format)

        records = self._get_records(
            mnemonic=mnemonic,
            starttime=starttime,
            endtime=endtime,
            time_format=time_format,
        )
        if len(records) == 0:
            return list()

        # If desired, remove bracket or outside of timeframe entries.
        if not include_bracket_values:
            selection = np.logical_and(
                records["MJD"] >= starttime.mjd, records["MJD"] <= endtime.mjd
            )
            records = records[selection]

        # Reformat to the desired list formatting.
        results = ValueCollection(
            include_obstime=include_obstime, zip_results=zip_results
        )
        values = records["EUValue"]
        obstimes = Time(records["MJD"], format="mjd")
        for obstime, value in zip(obstimes, values, strict=False):
            results.append(obstime, value)

        return results.collection

    def set_session(self):
        """Set up HTTP session."""
        self._req = requests.Request(
            method="GET",
            url=self.base_url + API_URI,
            headers={"Authorization": f"token {self.token}"},
        )

        s = requests.Session()
        retries = Retry(
            total=self.retries,
            backoff_factor=1.0,
            status_forcelist=FORCE_STATUSES,
            raise_on_status=True,
        )
        s.mount("https://", HTTPAdapter(max_retries=retries))

        self._session = s

    def _get_records(
        self, mnemonic, starttime, endtime, time_format=None, **other_kwargs
    ):
        """
        Retrieve all results for a mnemonic in the requested time range.

        Parameters
        ----------
        mnemonic : str
            The engineering mnemonic to retrieve

        starttime : str or astropy.time.Time
            The, inclusive, start time to retrieve from.

        endtime : str or astropy.time.Time
            The, inclusive, end time to retrieve from.

        time_format : str
            The format of the input time used if the input times
            are strings. If None, a guess is made.

        **other_kwargs : dict
            Keyword arguments not relevant to this implementation.

        Returns
        -------
        records : `astropy.Table`
            Returns the resulting table.

        Notes
        -----
        The engineering service always returns the bracketing entries
        before and after the requested time range.
        """
        if not isinstance(starttime, Time):
            starttime = Time(starttime, format=time_format)
        if not isinstance(endtime, Time):
            endtime = Time(endtime, format=time_format)
        self.starttime = starttime
        self.endtime = endtime

        # Make the query.
        response = self.mr.get_mnemonic_data(mnemonic, starttime.isot, endtime.isot)
        self.response = response.copy()

        # Convert to table.
        data = response["Data"]
        del response["Data"]
        table = Table(rows=data, meta=response)

        # Create a column MJD that has the MJD version of the data
        if len(table):
            obstime = Time(table["ObsTime"])
            table["MJD"] = obstime.mjd

        return table

    def __repr__(self):
        """What am I"""
        mr = self.mr
        db_config = mr.db_config.copy()
        del db_config["ad_name"]
        repr = f"{self.__class__.__name__}(config={db_config}, mission='{mr.mission}', env='{mr.env}')"
        return repr



"""
Module to calculate DAOFind sharpness and roundness1 properties.
"""

import warnings

import numpy as np
from astropy.convolution import Gaussian2DKernel
from astropy.nddata.utils import NoOverlapError, extract_array
from astropy.stats import gaussian_fwhm_to_sigma
from astropy.utils.decorators import lazyproperty
from scipy import ndimage


class DAOFindCatalog:
    """
    Class to calculate DAOFind sharpness and roundness1 properties.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        The input 2D image data.

    xypos : 2D `~numpy.ndarray`
        The x/y positions of the sources in the image.

    kernel_sigma : float
        The standard deviation of the Gaussian kernel used for
        convolution. The kernel size is determined by the kernel_sigma
        value.
    """

    def __init__(self, data, xypos, kernel_fwhm):
        self.data = data
        self.xypos = xypos
        self.kernel_sigma = kernel_fwhm * gaussian_fwhm_to_sigma

        self.names = ["sharpness", "roundness1"]

    @lazyproperty
    def kernel_size(self):
        """
        The DAOFind kernel size (in both x and y dimensions).
        """
        # always odd
        return 2 * int(max(2.0, 1.5 * self.kernel_sigma)) + 1

    @lazyproperty
    def kernel_center(self):
        """
        The DAOFind kernel x/y center.
        """
        return (self.kernel_size - 1) // 2

    @lazyproperty
    def kernel_mask(self):
        """
        The DAOFind kernel circular mask.

        NOTE: 1=good pixels, 0=masked pixels
        """
        yy, xx = np.mgrid[0 : self.kernel_size, 0 : self.kernel_size]
        radius = np.hypot(xx - self.kernel_center, yy - self.kernel_center)
        return (radius <= max(2.0, 1.5 * self.kernel_sigma)).astype(int)

    @lazyproperty
    def kernel(self):
        """
        The DAOFind kernel, a 2D circular Gaussian normalized to have
        zero sum.
        """
        size = self.kernel_size
        kernel = Gaussian2DKernel(self.kernel_sigma, x_size=size, y_size=size).array
        kernel *= self.kernel_mask
        kernel /= np.max(kernel)

        # normalize the kernel to zero sum
        npixels = self.kernel_mask.sum()
        denom = np.sum(kernel**2) - (np.sum(kernel) ** 2 / npixels)
        return ((kernel - (kernel.sum() / npixels)) / denom) * self.kernel_mask

    @lazyproperty
    def convolved_data(self):
        """
        The input data convolved with the DAOFind kernel.
        """
        return ndimage.convolve(self.data, self.kernel, mode="constant", cval=0.0)

    def _make_cutouts(self, array):
        """
        Make 2D cutouts centered on each source from the input data.

        Parameters
        ----------
        array : 2D `~numpy.ndarray`
            The input 2D image from which to make cutouts.

        Returns
        -------
        cutout : 3D `~numpy.ndarray`
            A 3D array containing 2D cutouts centered on each source from the
            input data. The cutout size always matches the size of the
            DAOFind kernel, which has odd dimensions.
        """
        cutouts = []
        for xcen, ycen in zip(*np.transpose(self.xypos), strict=True):
            try:
                cutout = extract_array(
                    array,
                    self.kernel.shape,
                    (ycen, xcen),
                    fill_value=0.0,
                )
            except NoOverlapError:
                cutout = np.zeros(self.kernel.shape)
            cutouts.append(cutout)

        return np.array(cutouts)  # all cutouts are the same size

    @lazyproperty
    def cutouts(self):
        """
        3D array containing 2D cutouts centered on each source from the
        input data.

        The cutout size always matches the size of the DAOFind kernel,
        which has odd dimensions.
        """
        return self._make_cutouts(self.data)

    @lazyproperty
    def cutouts_conv(self):
        """
        3D array containing 2D cutouts centered on each source from the
        input data convolved with the DAOFind kernel.

        The cutout size always matches the size of the DAOFind kernel,
        which has odd dimensions.
        """
        return self._make_cutouts(self.convolved_data)

    @lazyproperty
    def sharpness(self):
        """
        The DAOFind source sharpness statistic.

        The sharpness statistic measures the ratio of the difference
        between the height of the central pixel and the mean of the
        surrounding non-bad pixels to the height of the best fitting
        Gaussian function at that point (based on the convolved data).

        Stars generally have a "sharpness" between 0.2 and 1.0.
        """
        npixels = self.kernel_mask.sum() - 1  # exclude the peak pixel
        data_masked = self.cutouts * self.kernel_mask
        data_peak = self.cutouts[:, self.kernel_center, self.kernel_center]
        conv_peak = self.cutouts_conv[:, self.kernel_center, self.kernel_center]
        data_mean = (np.sum(data_masked, axis=(1, 2)) - data_peak) / npixels

        with warnings.catch_warnings():
            # ignore 0 / 0 for non-finite xypos
            warnings.simplefilter("ignore", category=RuntimeWarning)
            value = (data_peak - data_mean) / conv_peak
            return value.astype(np.float32)

    @lazyproperty
    def roundness1(self):
        """
        The DAOFind source roundness1 statistic based on symmetry.

        The roundness1 statistic computes the ratio of a measure of the
        bilateral symmetry of the object to a measure of the four-fold
        symmetry of the object.

        "Round" objects have a "roundness1" close to 0, generally
        between -1 and 1.
        """
        # set the central (peak) pixel to zero
        data = self.cutouts_conv.copy()
        data[:, self.kernel_center, self.kernel_center] = 0.0

        # calculate the four roundness quadrants
        quad1 = data[:, 0 : self.kernel_center + 1, self.kernel_center + 1 :]
        quad2 = data[:, 0 : self.kernel_center, 0 : self.kernel_center + 1]
        quad3 = data[:, self.kernel_center :, 0 : self.kernel_center]
        quad4 = data[:, self.kernel_center + 1 :, self.kernel_center :]

        axis = (1, 2)
        sum2 = (
            -quad1.sum(axis=axis)
            + quad2.sum(axis=axis)
            - quad3.sum(axis=axis)
            + quad4.sum(axis=axis)
        )
        sum2[sum2 == 0] = 0.0

        sum4 = np.abs(data).sum(axis=axis)
        sum4[sum4 == 0] = np.nan

        with warnings.catch_warnings():
            # ignore 0 / 0 for non-finite xypos
            warnings.simplefilter("ignore", category=RuntimeWarning)
            value = 2.0 * sum2 / sum4
            return value.astype(np.float32)



#!/usr/bin/env python

"""Pointing verification."""

import argparse
import logging
from sys import stdout

from romancal.orientation import pointing_summary


# Begin execution
def main():
    """Run pointing verification."""
    parser = argparse.ArgumentParser(
        description="""Summarize various pointing information in a table. Compare
                    the calculated V1 and REFPOINT pointing with the proposed TARGET pointing.
                    E.g.
                    $ pointing_summary exp1.fits
                    $ pointing_summary *.fits
                    """
    )

    parser.add_argument(
        "exposures", type=str, nargs="+", help="List of Roman data files to examine."
    )
    parser.add_argument(
        "-o",
        "--output",
        type=argparse.FileType("w"),
        default=stdout,
        help="File to write summary table to. Default is standard output.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase verbosity. Specifying multiple times adds more output.",
    )
    parser.add_argument(
        "--extra-meta",
        type=str,
        nargs="+",
        help="Extra meta information from the exposures to add to the result table",
    )

    args = parser.parse_args()

    # Set output detail.
    log_handler = logging.StreamHandler()
    logger = logging.getLogger("romancal")
    logger.addHandler(log_handler)
    LogLevels = [logging.WARNING, logging.INFO, logging.DEBUG]
    level = LogLevels[min(len(LogLevels) - 1, args.verbose)]
    logger.setLevel(level)

    # Process the file list.
    logger.info("Starting pointing summary.")
    deltas = pointing_summary.calc_deltas(args.exposures, extra_meta=args.extra_meta)
    deltas.write(args.output, format="ascii.ecsv")
    logger.info("........Summary completed.")


if __name__ == "__main__":
    main()



import asdf
from roman_datamodels import open as datamodels_open
from stpipe.library import AbstractModelLibrary, NoGroupID

from romancal.associations import AssociationNotValidError, load_asn

__all__ = ["ModelLibrary"]


class ModelLibrary(AbstractModelLibrary):
    @property
    def crds_observatory(self):
        return "roman"

    def _model_to_filename(self, model):
        model_filename = model.meta.filename
        if model_filename is None:
            model_filename = "model.asdf"
        return model_filename

    def _datamodels_open(self, filename, **kwargs):
        return datamodels_open(filename, **kwargs)

    @classmethod
    def _load_asn(cls, asn_path):
        try:
            with open(asn_path) as asn_file:
                asn_data = load_asn(asn_file)
        except AssociationNotValidError as e:
            raise OSError("Cannot read ASN file.") from e
        return asn_data

    def _filename_to_group_id(self, filename):
        """
        Compute a "group_id" without loading the file as a DataModel

        This function will return the meta.group_id stored in the ASDF
        extension (if it exists) or a group_id calculated from the
        ASDF headers.
        """
        meta = asdf.util.load_yaml(filename)["roman"]["meta"]
        if group_id := meta.get("group_id"):
            return group_id
        if observation_id := meta.get("observation", {}).get("observation_id", None):
            return observation_id
        raise NoGroupID(f"{filename} missing group_id")

    def _model_to_group_id(self, model):
        """
        Compute a "group_id" from a model using the DataModel interface
        """
        if (group_id := getattr(model.meta, "group_id", None)) is not None:
            return group_id
        if hasattr(model.meta, "observation") and hasattr(
            model.meta.observation, "observation_id"
        ):
            return model.meta.observation.observation_id
        raise NoGroupID(f"{model} missing group_id")

    def _assign_member_to_model(self, model, member):
        # roman_datamodels doesn't allow assignment of attributes
        # not defined in the schema. To work around this use
        # __setitem__ calls here instead of setattr
        for attr in ("tweakreg_catalog",):
            if attr in member:
                model.meta[attr] = member[attr]

        for asn_attr, dm_attr in (
            ("table_name", "table_name"),
            ("asn_pool", "pool_name"),
        ):
            if asn_attr not in self.asn:
                continue
            if not hasattr(model.meta, "asn"):
                model.meta["asn"] = {}
            model.meta.asn[dm_attr] = self.asn[asn_attr]



import logging

import numpy as np
from photutils.segmentation import SegmentationImage
from roman_datamodels.datamodels import (
    ForcedImageSourceCatalogModel,
    ImageModel,
    ImageSourceCatalogModel,
    MosaicSegmentationMapModel,
    SegmentationMapModel,
)
from roman_datamodels.stnode import SourceCatalog

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def save_segment_image(self, segment_img, source_catalog_model, output_filename):
    """
    Save the segmentation image to an ASDF file.

    Parameters
    ----------
    self : `romancal.source_catalog.SourceCatalogStep` or `romancal.multiband_catalog.MultiBandCatalogStep`
        The step instance that is calling this function.

    segment_img : `photutils.segmentation.SegmentationImage`
        The segmentation image to save.

    source_catalog_model : `roman_datamodels.datamodels.ImageSourceCatalogModel` or `roman_datamodels.datamodels.MosaicSourceCatalogModel` or `roman_datamodels.datamodels.ForcedImageSourceCatalogModel` or `roman_datamodels.datamodels.ForcedMosaicSourceCatalogModel`
        The source catalog model to use for metadata.

    output_filename : str
        The output file name.
    """
    if segment_img is None:
        log.warning("No segmentation image to save.")
        return

    # Define the segmentation model based on the source catalog model
    # type
    if isinstance(
        source_catalog_model, (ImageSourceCatalogModel | ForcedImageSourceCatalogModel)
    ):
        segm_model = SegmentationMapModel
    else:
        segm_model = MosaicSegmentationMapModel
    segmentation_model = segm_model.create_minimal({"meta": source_catalog_model.meta})

    # Set the data and detection image
    segmentation_model.data = segment_img.data.astype(np.uint32)
    if hasattr(segment_img, "detection_image"):
        segmentation_model["detection_image"] = segment_img.detection_image

    # Save the segmentation image to the output file
    self.output_ext = "asdf"
    self.save_model(
        segmentation_model,
        output_file=output_filename,
        suffix="segm",
        force=True,
    )


def save_all_results(self, segment_img, cat_model, input_model=None):
    """
    Return and save the results of the source catalog step.

    The segmentation image is always saved.

    If ``save_results`` is True, then either the input model with
    updated metadata or the source catalog model is saved/returned
    depending on the value of ``return_updated_model``. If
    ``return_updated_model`` is True, then the input model is
    saved/returned with updated metadata. If False, then the source
    catalog model is saved/returned.

    The source catalog is saved as a parquet file.

    Parameters
    ----------
    self : `romancal.source_catalog.SourceCatalogStep` or `romancal.multiband_catalog.MultiBandCatalogStep`
        The step instance that is calling this function.

    segment_img : `photutils.segmentation.SegmentationImage`
        The segmentation image created by the source catalog step.

    cat_model : `datamodels.ImageSourceCatalogModel`, `datamodels.MosaicSourceCatalogModel`, `datamodels.ForcedImageSourceCatalogModel`, or `datamodels.ForcedMosaicSourceCatalogModel`
        The source catalog model created by the step

    input_model : `None`, `ImageModel`, or `MosaicModel`, optional
        The input model to the source catalog step. This
        is required only for the SourceCatalogStep when
        ``self.return_updated_model`` is `True`.

    Returns
    -------
    result : `ImageModel`, `MosaicModel`, or `SourceCatalog`
        The source catalog model or the input model with updated
        metadata.
    """
    if isinstance(segment_img, np.ndarray):
        # convert the segmentation image to a SegmentationImage
        segment_img = SegmentationImage(segment_img)

    # define the output filename
    output_filename = (
        self.output_file if self.output_file is not None else cat_model.meta.filename
    )

    # always save the segmentation image
    save_segment_image(self, segment_img, cat_model, output_filename)

    # Update the source catalog filename metadata
    self.output_ext = "parquet"
    output_catalog_name = self.make_output_path(basepath=output_filename, suffix="cat")
    self.output_ext = "asdf"

    # Always save the source catalog, but don't save it twice.
    # If save_results=False or return_update_model=True, we need to
    # explicitly save it.
    return_updated_model = getattr(self, "return_updated_model", False)
    if not self.save_results or return_updated_model:
        self.output_ext = "parquet"
        self.save_model(
            cat_model,
            output_file=output_filename,
            suffix="cat",
            force=True,
        )
        self.output_ext = "asdf"

    # Return the source catalog object or the input model. If the
    # input model is an ImageModel, the metadata is updated with the
    # source catalog filename.
    if return_updated_model:
        # define the catalog filename; self.save_model will
        # determine whether to use a fully qualified path

        # set the suffix to something else to prevent the step from
        # overwriting the source catalog file with a datamodel
        self.suffix = "sourcecatalog"

        # update the input datamodel with the tweakreg catalog name
        if isinstance(input_model, ImageModel):
            input_model.meta.source_catalog = SourceCatalog()
            input_model.meta.source_catalog.tweakreg_catalog_name = output_catalog_name
            input_model.meta.cal_step.source_catalog = "COMPLETE"

        result = input_model
    else:
        self.output_ext = "parquet"
        result = cat_model

    if self.output_file is None:
        self.output_file = cat_model.meta.filename

    # validate the result to flush out any lazy-loaded contents
    result.validate()
    return result


def save_empty_results(self, segment_shape, cat_model, input_model=None, msg=""):
    """
    Return and save empty results for the source catalog or multiband
    catalog step.

    Parameters
    ----------
    self : `romancal.source_catalog.SourceCatalogStep` or `romancal.multiband_catalog.MultiBandCatalogStep`
        The step instance that is calling this function.

    segment_shape : tuple
        The shape of the segmentation image to create.

    cat_model : `datamodels.ImageSourceCatalogModel`, `datamodels.MosaicSourceCatalogModel`, `datamodels.ForcedImageSourceCatalogModel`, or `datamodels.ForcedMosaicSourceCatalogModel`
        The source catalog model created by the step.

    input_model : `None`, `ImageModel`, or `MosaicModel`, optional
        The input model to the source catalog step. This
        is required only for the SourceCatalogStep when
        ``self.return_updated_model`` is `True`.

    msg : str, optional
        A message to log if no results were found. Default is an empty
        string.

    Returns
    -------
    result : `ImageModel`, `MosaicModel`, or `SourceCatalog`
        The source catalog model or the input model with updated
        metadata.
    """
    if msg:
        log.error(msg)

    segment_img = np.zeros(segment_shape, dtype=np.uint32)
    cat_model.source_catalog = cat_model.create_empty_catalog()
    return save_all_results(self, segment_img, cat_model, input_model=input_model)



"""
convert old patch names (i.e. `r274dp63x31y80`) from the old patch table
to the names of the closest sky cells (i.e. `270p65x48y69`) from the new sky cell table

remove this script when we no longer have any old files with old patch names
"""

from pathlib import Path

import asdf
import numpy as np
import spherical_geometry.vector as sgv

import romancal.skycell.skymap as sc

PATCH_TABLE_PATH = Path("/grp/roman/scsb/tesselation/patches.asdf")


def patch_names_to_skycell_names(
    patch_names: list[str],
    patch_table_path: Path = PATCH_TABLE_PATH,
    skymap: sc.SkyMap = sc.SKYMAP,
) -> list[str]:
    """
    convert old patch names (i.e. `r274dp63x31y80`) from the old patch table
    to the names of the closest sky cells (i.e. `270p65x48y69`) from the new sky cell table

    Parameters
    ----------
    patch_names : list[str]
        list of patch names, i.e. `r274dp63x31y80`
    patch_table_path : Path
        path to patch table; defaults to /grp/roman/scsb/tesselation/patches.asdf
    skymap : romancal.skycell.skymap.SkyMap
        SkyMap instance; defaults to global `skycells` table from CRDS

    Returns
    -------
    list of sky cell names, i.e. `270p65x48y69`

    Examples
    --------
    >>> patch_names_to_skycell_names(["r274dp63x31y80", "r274dp63x31y81", "r274dp63x32y82", "r274dp63x32y80", "r274dp63x32y81", "r274dp63x32y82"])
    ['225p90x49y67', '225p90x49y69', '225p90x50y70', '225p90x50y67', '225p90x50y69', '225p90x50y70']
    """

    patch_table = asdf.open(patch_table_path, memmap=True)

    skycell_names = []
    projregions = {}
    for patch_name in patch_names:
        patch = patch_table["patches"][
            np.where(patch_table["patches"]["name"] == patch_name)
        ]
        patch_radec_center = np.array(
            sgv.lonlat_to_vector(patch["ra_center"], patch["dec_center"])
        ).T[0]
        neighboring_projregion_indices = skymap.projection_regions_kdtree.query(
            patch_radec_center,
            k=4,
        )[1]

        skycell_indices = {}
        for index in neighboring_projregion_indices:
            if index not in projregions:
                projregions[index] = sc.ProjectionRegion(index)
            distance, skycell_index = projregions[index].skycells_kdtree.query(
                patch_radec_center, k=1
            )
            skycell_indices[distance] = skycell_index

        skycell_names.append(sc.SkyCell(skycell_indices[min(skycell_indices)]).name)

    return np.array(skycell_names).tolist()



"""
Unit tests for the Roman source detection step code
"""

from copy import deepcopy

import crds
import numpy as np
import pytest
import roman_datamodels as rdm
from astropy import units as u
from astropy.modeling.models import Gaussian2D
from astropy.stats import mad_std
from astropy.table import QTable
from photutils.datasets import make_model_image
from photutils.psf import PSFPhotometry
from roman_datamodels import stnode
from roman_datamodels.datamodels import ImageModel

from romancal.source_catalog.psf import (
    azimuthally_smooth,
    fit_psf_to_image_model,
    get_gridded_psf_model,
)

n_trials = 15
image_model_shape = (50, 50)
rng = np.random.default_rng(0)


def setup_inputs(
    shape=image_model_shape,
    noise=1.0,
    seed=None,
):
    """
    Return ImageModel of level 2 image.
    """
    wfi_image = stnode.WfiImage.create_fake_data(shape=shape)
    wfi_image.data = np.ones(shape, dtype=np.float32)
    wfi_image.meta.filename = "filename"
    wfi_image.meta.instrument["optical_element"] = "F087"

    # add noise to data
    if noise is not None:
        setup_rng = np.random.default_rng(seed or 19)
        wfi_image.data = setup_rng.normal(scale=noise, size=shape).astype("float32")
        wfi_image.err = noise * (np.ones(shape, dtype=np.float32) * u.DN / u.s).value

    # add dq array
    wfi_image.dq = np.zeros(shape, dtype=np.uint32)

    # construct ImageModel
    mod = ImageModel(wfi_image)

    crds_parameters = mod.get_crds_parameters()
    crds_ref_file = crds.getreferences(
        crds_parameters, reftypes=["epsf"], observatory="roman"
    )
    psf_ref_file = crds_ref_file["epsf"]
    mod["psf_ref_model"] = rdm.open(psf_ref_file)

    # compute gridded PSF model:
    psf_model = get_gridded_psf_model(mod["psf_ref_model"])

    return mod, psf_model


def add_sources(image_model, psf_model, x_true, y_true, flux_true, background=10):
    params_table = QTable()
    params_table["x_0"] = np.atleast_1d(x_true)
    params_table["y_0"] = np.atleast_1d(y_true)
    params_table["flux"] = np.atleast_1d(flux_true)

    shape = image_model.data.shape
    image = make_model_image(shape, psf_model, params_table, model_shape=(19, 19))
    image += rng.normal(background, image_model.err, size=shape)

    image_model.data = image * np.ones_like(image_model.data)


class TestPSFFitting:
    def setup_method(self):
        self.image_model, self.psf_model = setup_inputs(
            shape=image_model_shape,
        )

    @pytest.mark.parametrize(
        "dx, dy, true_flux",
        zip(
            rng.uniform(-1, 1, n_trials),
            rng.uniform(-1, 1, n_trials),
            np.geomspace(1_000, 100_000, n_trials),
            strict=False,
        ),
    )
    def test_psf_fit(self, dx, dy, true_flux):
        # generate an ImageModel
        image_model = deepcopy(self.image_model)

        # add synthetic sources to the ImageModel:
        true_x = image_model_shape[0] / 2 + dx
        true_y = image_model_shape[1] / 2 + dy
        add_sources(image_model, self.psf_model, true_x, true_y, true_flux)
        init_data_stddev = np.std(image_model.data)

        # fit the PSF to the ImageModel:
        results_table, photometry = fit_psf_to_image_model(
            image_model=image_model,
            photometry_cls=PSFPhotometry,
            psf_model=self.psf_model,
            x_init=true_x,
            y_init=true_y,
        )

        # difference between input and output, normalized by the
        # uncertainty. Has units of sigma:
        delta_x = np.abs(true_x - results_table["x_fit"]) / results_table["x_err"]
        delta_y = np.abs(true_y - results_table["y_fit"]) / results_table["y_err"]

        sigma_threshold = 3.5
        assert np.all(delta_x < sigma_threshold)
        assert np.all(delta_y < sigma_threshold)

        # now check that the uncertainties aren't way too large, which could cause
        # the above test to pass even when the fits are bad. Use overly-simple
        # approximation that astrometric uncertainty be proportional to the
        # PSF's FWHM / SNR:
        approx_snr = true_flux / init_data_stddev
        approx_fwhm = 1
        approx_centroid_err = approx_fwhm / approx_snr

        # centroid err heuristic above is an underestimate, so we scale it up:
        scale_factor_approx = 2

        assert np.all(
            results_table["x_err"] < scale_factor_approx * approx_centroid_err
        )
        assert np.all(
            results_table["y_err"] < scale_factor_approx * approx_centroid_err
        )


def test_azimuthally_smooth():
    """Test azimuthally smoothing"""
    grid_x, grid_y = np.mgrid[0:201, 0:201]
    gauss_model = Gaussian2D(1.0, 100, 100, 20, 20)
    gauss = gauss_model(grid_x, grid_y)
    smoothed = azimuthally_smooth(gauss, oversample=1)
    delta = gauss - smoothed

    assert np.mean(delta) < 1.0e-6
    assert mad_std(delta) < 1.0e-8



import os
from pathlib import Path

import astropy.units as u
import numpy as np
import pyarrow
import pytest
from astropy.modeling.models import Gaussian2D
from astropy.table import Table
from astropy.time import Time
from numpy.testing import assert_equal
from photutils.segmentation import SegmentationImage
from roman_datamodels import datamodels as rdm
from roman_datamodels import stnode
from roman_datamodels.datamodels import (
    ForcedImageSourceCatalogModel,
    ImageModel,
    ImageSourceCatalogModel,
    MosaicModel,
    MosaicSegmentationMapModel,
    MosaicSourceCatalogModel,
    SegmentationMapModel,
)

from romancal.source_catalog.source_catalog import RomanSourceCatalog
from romancal.source_catalog.source_catalog_step import SourceCatalogStep


def make_test_image():
    g1 = Gaussian2D(121.0, 11.1, 12.2, 1.5, 1.5)
    g2 = Gaussian2D(70, 65, 18, 9.2, 4.5)
    g3 = Gaussian2D(111.0, 41, 42.7, 8.0, 3.0, theta=30 * u.deg)
    g4 = Gaussian2D(81.0, 17, 52.7, 4, 2, theta=102 * u.deg)
    g5 = Gaussian2D(107.0, 65, 71, 12, 2, theta=142 * u.deg)
    g6 = Gaussian2D(50, 20, 80, 2.1, 2.1)
    g7 = Gaussian2D(97.0, 85, 87.3, 4, 2, theta=-30 * u.deg)

    yy, xx = np.mgrid[0:101, 0:101]
    data = (
        g1(xx, yy)
        + g2(xx, yy)
        + g3(xx, yy)
        + g4(xx, yy)
        + g5(xx, yy)
        + g6(xx, yy)
        + g7(xx, yy)
    ).value.astype("float32")

    y0 = 2
    x0 = 90
    dd = 5
    value = -20
    y1 = y0 + dd
    x1 = x0 + dd - 1
    x2 = x0 + dd
    x3 = x0 + 2 * dd - 1
    data[y0:y1, x0] = value
    data[y0, x0:x1] = value
    data[y0:y1, x2] = value
    data[y0, x2:x3] = value
    data[y0 + dd // 2, x2:x3] = value
    data[y1 - 1, x2:x3] = value
    data[y0 + 1 : y1 - 1, x3] = value

    rng = np.random.default_rng(seed=123)
    noise_scale = 2.5
    noise = rng.normal(0, noise_scale, size=data.shape)
    data += noise
    err = np.zeros_like(data) + noise_scale

    return data, err


@pytest.fixture
def mosaic_model():
    defaults = {
        "meta": {
            "coadd_info": {
                "time_first": Time("2024-01-01T12:00:00.000", format="isot"),
            },
            "instrument": {
                "optical_element": "F158",
            },
            "resample": {"pixfrac": 1.0},
            "wcsinfo": {
                "pixel_scale": 1.5277777769528157e-05
            },  # Taken from regtest test L3 mosaic.
        }
    }
    model = MosaicModel.create_fake_data(defaults=defaults, shape=(101, 101))
    model.meta.ref_file = stnode.RefFile.create_fake_data()
    model.meta.filename = "none"
    model.meta.cal_step = stnode.L3CalStep.create_fake_data()
    model.cal_logs = stnode.CalLogs.create_fake_data()
    data, err = make_test_image()
    model.data = data
    model.err = err
    model.weight = 1.0 / err
    return model


@pytest.fixture
def image_model():
    model = ImageModel.create_fake_data(shape=(101, 101))
    model.meta.filename = "none"
    model.meta.cal_step = stnode.L2CalStep.create_fake_data()
    model.meta.cal_logs = stnode.CalLogs.create_fake_data()
    data, err = make_test_image()
    model.data = data
    model.err = err
    model.meta.photometry.conversion_megajanskys = (0.3324 * u.MJy / u.sr).value
    return model


def test_forced_catalog(image_model, tmp_path):
    os.chdir(tmp_path)
    step = SourceCatalogStep()
    _ = step.call(
        image_model,
        bkg_boxsize=50,
        kernel_fwhm=2.0,
        snr_threshold=5,
        npixels=10,
        save_results=True,
        output_file="source_cat.asdf",
    )
    result_force = step.call(
        image_model,
        bkg_boxsize=50,
        kernel_fwhm=2.0,
        snr_threshold=5,
        npixels=10,
        save_results=True,
        output_file="force_cat.asdf",
        forced_segmentation="source_segm.asdf",
    )
    assert isinstance(result_force, ForcedImageSourceCatalogModel)
    catalog = result_force.source_catalog
    assert isinstance(catalog, Table)
    has_forced_fields = False
    for field in catalog.dtype.names:
        if "forced_" in field:
            has_forced_fields = True
    assert has_forced_fields


@pytest.mark.parametrize(
    "snr_threshold, npixels, nsources, save_results",
    (
        (3, 10, 7, True),
        (3, 50, 5, False),
        (10, 10, 7, False),
        (20, 10, 5, False),
        (25, 10, 3, False),
        (35, 10, 1, False),
        (50, 10, 0, False),
    ),
)
def test_l2_source_catalog(
    image_model, snr_threshold, npixels, nsources, save_results, tmp_path
):
    os.chdir(tmp_path)
    step = SourceCatalogStep()
    result = step.call(
        image_model,
        bkg_boxsize=50,
        kernel_fwhm=2.0,
        snr_threshold=snr_threshold,
        npixels=npixels,
        save_results=save_results,
    )

    cat = result.source_catalog
    assert isinstance(cat, Table)
    assert len(cat) == nsources

    if nsources > 0:
        for colname in cat.colnames:
            if (
                "flux" in colname
                and "fluxfrac" not in colname
                and "aper_bkg_flux" not in colname
            ):
                assert cat[colname].unit == "nJy"
        assert np.min(cat["x_centroid"]) > 0.0
        assert np.min(cat["y_centroid"]) > 0.0
        assert np.max(cat["x_centroid"]) < 100.0
        assert np.max(cat["y_centroid"]) < 100.0
        assert np.any(cat["ra"])
        assert np.any(cat["dec"])


@pytest.mark.parametrize(
    "snr_threshold, npixels, nsources, save_results",
    (
        (3, 10, 7, True),
        (3, 50, 5, False),
        (10, 10, 7, False),
        (20, 10, 5, False),
        (25, 10, 3, False),
        (35, 10, 1, False),
        (50, 10, 0, False),
    ),
)
def test_l3_source_catalog(
    mosaic_model, snr_threshold, npixels, nsources, save_results, tmp_path
):
    os.chdir(tmp_path)
    step = SourceCatalogStep()

    # Create model and set some crucial meta required to
    # create the L3 PSF for flux determination.
    result = step.call(
        mosaic_model,
        bkg_boxsize=50,
        kernel_fwhm=2.0,
        snr_threshold=snr_threshold,
        npixels=npixels,
        save_results=save_results,
    )

    cat = result.source_catalog
    assert isinstance(cat, Table)
    assert len(cat) == nsources

    if nsources > 0:
        for colname in cat.colnames:
            if (
                "flux" in colname
                and "fluxfrac" not in colname
                and "aper_bkg_flux" not in colname
            ):
                assert cat[colname].unit == "nJy"
        assert np.min(cat["x_centroid"]) > 0.0
        assert np.min(cat["y_centroid"]) > 0.0
        assert np.max(cat["x_centroid"]) < 100.0
        assert np.max(cat["y_centroid"]) < 100.0
        assert np.any(cat["ra"])
        assert np.any(cat["dec"])


def test_background(mosaic_model, tmp_path):
    """
    Test background fallback when Background2D fails.
    """
    os.chdir(tmp_path)
    step = SourceCatalogStep()
    result = step.call(
        mosaic_model,
        bkg_boxsize=1000,
        kernel_fwhm=2.0,
        snr_threshold=3,
        npixels=25,
        fit_psf=False,
    )

    cat = result.source_catalog

    assert isinstance(cat, Table)


def test_l2_input_model_unchanged(image_model, tmp_path):
    """
    Test that the input model data and error arrays are unchanged after
    processing by SourceCatalogStep.
    """
    os.chdir(tmp_path)
    original_data = image_model.data.copy()
    original_err = image_model.err.copy()

    step = SourceCatalogStep()
    step.call(
        image_model,
        snr_threshold=0.5,
        npixels=5,
        bkg_boxsize=50,
        kernel_fwhm=2.0,
        save_results=False,
        fit_psf=False,
    )

    assert_equal(original_data, image_model.data)
    assert_equal(original_err, image_model.err)


def test_l3_input_model_unchanged(mosaic_model, tmp_path):
    """
    Test that the input model data and error arrays are unchanged after
    processing by SourceCatalogStep.
    """
    os.chdir(tmp_path)
    original_data = mosaic_model.data.copy()
    original_err = mosaic_model.err.copy()

    step = SourceCatalogStep()
    step.call(
        mosaic_model,
        snr_threshold=0.5,
        npixels=5,
        bkg_boxsize=50,
        kernel_fwhm=2.0,
        save_results=False,
        fit_psf=False,
    )

    assert_equal(original_data, mosaic_model.data)
    assert_equal(original_err, mosaic_model.err)


def test_invalid_step_inputs(image_model, mosaic_model):
    for input_model in (image_model, mosaic_model):
        model = input_model.copy()
        model.data = np.full(model.data.shape, np.nan)
        step = SourceCatalogStep()
        result = step.call(model)
        cat = result.source_catalog
        assert isinstance(cat, Table)
        assert len(cat) == 0


def test_inputs(mosaic_model):
    data = np.ones((3, 3), dtype=int)
    data[1, 1] = 1
    segm = SegmentationImage(data)
    cdata = np.ones((3, 3))
    kernel_fwhm = 2.0
    with pytest.raises(ValueError):
        RomanSourceCatalog(np.ones((3, 3)), segm, cdata, kernel_fwhm, fit_psf=True)


def test_psf_photometry(tmp_path, image_model):
    """
    Test PSF photometry.
    """
    step = SourceCatalogStep()
    result = step.call(
        image_model,
        bkg_boxsize=20,
        kernel_fwhm=2.0,
        snr_threshold=3,
        npixels=10,
        save_results=False,
    )

    cat = result.source_catalog
    assert isinstance(cat, Table)
    assert len(cat) == 7

    for colname in cat.colnames:
        if (
            "flux" in colname
            and "fluxfrac" not in colname
            and "aper_bkg_flux" not in colname
        ):
            assert cat[colname].unit == "nJy"

    for colname in cat.colnames:
        if "psf" in colname:
            assert len(cat[colname])  # make sure the column isn't empty
            assert not np.any(np.isnan(cat[colname]))  # and contains no nans


@pytest.mark.parametrize("fit_psf", [True, False])
def test_do_psf_photometry_column_names(tmp_path, image_model, fit_psf):
    """
    Test that fit_psf will determine whether the PSF
    photometry columns are added to the final catalog or not.
    """
    step = SourceCatalogStep()
    result = step.call(
        image_model,
        bkg_boxsize=20,
        kernel_fwhm=2.0,
        snr_threshold=3,
        npixels=10,
        save_results=False,
        fit_psf=fit_psf,
    )

    cat = result.source_catalog
    assert isinstance(cat, Table)

    psf_colnames = []
    for colname in cat.colnames:
        if "psf" in colname:
            psf_colnames.append(colname)

    if fit_psf:
        assert len(psf_colnames) > 0
    else:
        assert len(psf_colnames) == 0


@pytest.mark.parametrize(
    "snr_threshold, npixels, nsources, save_results, return_updated_model, expected_result, expected_outputs",
    (
        (
            3,
            10,
            7,
            True,
            True,
            ImageModel,
            {
                "cat": ImageSourceCatalogModel,
                "segm": SegmentationMapModel,
                "sourcecatalog": ImageModel,
            },
        ),
        (
            3,
            50,
            5,
            True,
            False,
            ImageSourceCatalogModel,
            {
                "cat": ImageSourceCatalogModel,
                "segm": SegmentationMapModel,
            },
        ),
        (
            10,
            10,
            7,
            False,
            True,
            ImageModel,
            {
                "cat": ImageSourceCatalogModel,
                "segm": SegmentationMapModel,
            },
        ),
        (
            20,
            10,
            5,
            False,
            False,
            ImageSourceCatalogModel,
            {
                "cat": ImageSourceCatalogModel,
                "segm": SegmentationMapModel,
            },
        ),
    ),
)
def test_l2_source_catalog_keywords(
    image_model,
    snr_threshold,
    npixels,
    nsources,
    save_results,
    return_updated_model,
    expected_result,
    expected_outputs,
    tmp_path,
    monkeypatch,
):
    """
    Test that the proper object is returned in the call to SourceCatalogStep
    and that the desired output files are saved to the disk with the correct type.
    """
    os.chdir(tmp_path)
    monkeypatch.setattr(
        SourceCatalogStep, "return_updated_model", return_updated_model, raising=False
    )

    result = SourceCatalogStep.call(
        image_model,
        bkg_boxsize=50,
        kernel_fwhm=2.0,
        snr_threshold=snr_threshold,
        npixels=npixels,
        save_results=save_results,
    )

    # assert that we returned the correct object
    assert isinstance(result, expected_result)

    # assert that the desired output files were saved to disk and that
    # they are of the correct type
    for suffix in expected_outputs.keys():
        if suffix == "cat":
            ext = "parquet"
        else:
            ext = "asdf"

        # annoying case.  Sometimes we have meta.filename as just "none" and
        # this test relies on the filename actually being at none_cat.parquet, etc.
        # But if we return a source catalog with a correct meta.filename (e.g.,
        # none_cat.parquet), this test needs to know how to translate that back
        # to the equivalent segmentation file.
        basefilename = result.meta.filename.split("_")[0]
        filepath = Path(tmp_path / f"{basefilename}_{suffix}.{ext}")
        assert filepath.exists()

        if suffix == "cat":
            # the catalog is saved as a parquet file
            tbl = pyarrow.parquet.read_table(filepath)
            assert isinstance(tbl, pyarrow.Table)
        else:
            assert isinstance(rdm.open(filepath), expected_outputs.get(suffix))


@pytest.mark.parametrize(
    "snr_threshold, npixels, nsources, save_results, return_updated_model, expected_result, expected_outputs",
    (
        (
            3,
            10,
            7,
            True,
            True,
            MosaicModel,
            {
                "cat": MosaicSourceCatalogModel,
                "segm": MosaicSegmentationMapModel,
                "sourcecatalog": MosaicModel,
            },
        ),
        (
            3,
            50,
            5,
            True,
            False,
            MosaicSourceCatalogModel,
            {
                "cat": MosaicSourceCatalogModel,
                "segm": MosaicSegmentationMapModel,
            },
        ),
        (
            10,
            10,
            7,
            False,
            True,
            MosaicModel,
            {
                "cat": MosaicSourceCatalogModel,
                "segm": MosaicSegmentationMapModel,
            },
        ),
        (
            20,
            10,
            5,
            False,
            False,
            MosaicSourceCatalogModel,
            {
                "cat": MosaicSourceCatalogModel,
                "segm": MosaicSegmentationMapModel,
            },
        ),
    ),
)
def test_l3_source_catalog_keywords(
    mosaic_model,
    snr_threshold,
    npixels,
    nsources,
    save_results,
    return_updated_model,
    expected_result,
    expected_outputs,
    tmp_path,
    monkeypatch,
):
    """
    Test that the proper object is returned in the call to SourceCatalogStep
    and that the desired output files are saved to the disk with the correct type.
    """
    os.chdir(tmp_path)
    # this step attribute controls whether to return a datamodel or source catalog
    monkeypatch.setattr(
        SourceCatalogStep, "return_updated_model", return_updated_model, raising=False
    )

    result = SourceCatalogStep.call(
        mosaic_model,
        bkg_boxsize=50,
        kernel_fwhm=2.0,
        snr_threshold=snr_threshold,
        npixels=npixels,
        save_results=save_results,
    )

    # assert that we returned the correct object
    assert isinstance(result, expected_result)

    # assert that the desired output files were saved to disk and that
    # they are of the correct type
    for suffix in expected_outputs.keys():
        if suffix == "cat":
            ext = "parquet"
        else:
            ext = "asdf"

        basefilename = result.meta.filename.split("_")[0]
        filepath = Path(tmp_path / f"{basefilename}_{suffix}.{ext}")
        assert filepath.exists()

        if suffix == "cat":
            # the catalog is saved as a parquet file
            tbl = pyarrow.parquet.read_table(filepath)
            assert isinstance(tbl, pyarrow.Table)
        else:
            assert isinstance(rdm.open(filepath), expected_outputs.get(suffix))


@pytest.mark.parametrize(
    "return_updated_model, expected_result",
    (
        (
            True,
            ImageModel,
        ),
        (
            False,
            ImageSourceCatalogModel,
        ),
    ),
)
def test_l2_source_catalog_return_updated_model_attribute(
    image_model,
    return_updated_model,
    expected_result,
    tmp_path,
):
    """
    Test that the proper object is returned in the call to SourceCatalogStep.
    """
    os.chdir(tmp_path)

    step = SourceCatalogStep(
        bkg_boxsize=50,
        kernel_fwhm=2.0,
        snr_threshold=3,
        npixels=10,
    )

    if return_updated_model:
        # mimic what happens in the ELP -- i.e. set the "hidden" parameter
        # to cause this step to return a model instead of a catalog
        step.return_updated_model = return_updated_model

    result = step.run(image_model)

    # assert that we returned the correct object
    assert isinstance(result, expected_result)



"""
Module to calculate the segmentation-based properties.
"""

import inspect
import warnings

import astropy.units as u
import numpy as np
from astropy.utils import lazyproperty
from photutils.segmentation import SourceCatalog


class SegmentCatalog:
    """
    Class to calculate the segmentation-based properties.

    Parameters
    ----------
    model : `ImageModel` or `MosaicModel`
        The input data model. The image data is assumed to be background
        subtracted.

    segment_image : `~photutils.segmentation.SegmentationImage`
        A 2D segmentation image, with the same shape as the input data,
        where sources are marked by different positive integer values. A
        value of zero is reserved for the background.

    convolved_data : 2D `~numpy.ndarray` or `None`
        The 2D array used to calculate the source centroid and shape
        measurements. The image is assumed to be background subtracted.

    pixel_area : `~astropy.units.Quantity`
        The pixel area in steradians. This is used to convert various
        measuments from pixels to arcseconds.

    wcs_angle : `~astropy.units.Quantity`
        The angle (as a Quantity in degrees) measured counterclockwise
        from the positive x axis to the "North" axis of the celestial
        coordinate system.

    detection_cat : `None` or `~photutils.segmentation.SourceCatalog`, optional
        A `~photutils.segmentation.SourceCatalog` object for the
        detection image. The segmentation image used to create
        the detection catalog must be the same one input to
        ``segment_image``. If input, then the detection catalog source
        centroids and morphological/shape properties will be returned
        instead of calculating them from the input ``data``. The
        detection catalog centroids and shape properties will also be
        used to perform aperture photometry (i.e., circular and Kron).

    flux_unit : str, optional
        The unit of the flux density. Default is 'nJy'.

    Notes
    -----
    ``model.err`` is assumed to be the total error array corresponding
    to the input science ``model.data`` array. It is assumed to include
    *all* sources of error, including the Poisson error of the sources,
    and have the same shape and units as the science data array.
    """

    def __init__(
        self,
        model,
        segment_img,
        convolved_data,
        pixel_area,
        wcs_angle,
        detection_cat=None,
        flux_unit="nJy",
    ):
        self.model = model
        self.segment_img = segment_img
        self.convolved_data = convolved_data
        self.pixel_area = pixel_area
        self.wcs_angle = wcs_angle
        self.detection_cat = detection_cat
        self.flux_unit = flux_unit

        self.names = []
        self.wcs = self.model.meta.wcs
        self.meta = {}

        # needed for detection_cat
        self.source_cat = None

        # calculate the segment properties
        self.calc_segment_properties()

        # lazyproperties are not set until accessed so we need to
        # manually append them
        for name in self._lazyproperties:
            self.names.append(name)

        # add the placeholder attributes
        self.add_placeholders()

    @property
    def _lazyproperties(self):
        """
        A list of all class lazyproperties.
        """

        def islazyproperty(obj):
            return isinstance(obj, lazyproperty)

        return [
            i[0] for i in inspect.getmembers(self.__class__, predicate=islazyproperty)
        ]

    @staticmethod
    def convert_flux_to_abmag(flux, flux_err):
        """
        Convert flux (and error) to AB magnitude (and error).

        Parameters
        ----------
        flux, flux_err : `~numpy.ndarray`
            The input flux and error arrays.

        Returns
        -------
        abmag, abmag_err : `~astropy.ndarray`
            The output AB magnitude and error arrays.
        """
        # ignore RunTimeWarning if flux or flux_err contains NaNs
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)

            abmag = flux.to(u.ABmag)
            abmag_err = 2.5 * np.log10(1.0 + (flux_err / flux))
            abmag_err = abmag_err.value * u.ABmag

            # handle negative fluxes
            idx = flux.value < 0
            abmag[idx] = np.nan
            abmag_err[idx] = np.nan

        return abmag, abmag_err

    @lazyproperty
    def pixel_scale(self):
        """
        The pixel scale in arcseconds.
        """
        # assumes square pixels
        return np.sqrt(self.pixel_area).to(u.arcsec)

    def calc_segment_properties(self):
        """
        Calculate the segment-based properties provided by
        `~photutils.segmentation.SourceCatalog`.

        The results are set as dynamic attributes on the class instance.
        """
        # detection_cat is a RomanSourceCatalog object
        # segment_cat is a SegmentCatalog object
        # source_cat is a SourceCatalog object
        if self.detection_cat is not None:
            detection_cat = self.detection_cat.segment_cat.source_cat
        else:
            detection_cat = None

        segm_cat = SourceCatalog(
            self.model.data,
            self.segment_img,
            convolved_data=self.convolved_data,
            error=self.model.err,
            wcs=self.wcs,
            detection_cat=detection_cat,
        )
        self.source_cat = segm_cat
        self.meta.update(segm_cat.meta)

        # Extract the properties from the segment catalog. These
        # names are the SourceCatalog property names and the order
        # is not important.
        photutils_names = (
            "label",
            "xcentroid",
            "ycentroid",
            "xcentroid_win",
            "ycentroid_win",
            "sky_centroid",
            "sky_centroid_win",
            "bbox_xmin",
            "bbox_xmax",
            "bbox_ymin",
            "bbox_ymax",
            "area",
            "semimajor_sigma",
            "semiminor_sigma",
            "fwhm",
            "orientation",
            "ellipticity",
            "cxx",
            "cxy",
            "cyy",
            "kron_radius",
            "segment_flux",
            "segment_fluxerr",
            "kron_flux",
            "kron_fluxerr",
        )

        # if needed, map names from photutils to the output catalog names
        name_map = {}
        name_map["xcentroid"] = "x_centroid"
        name_map["ycentroid"] = "y_centroid"
        name_map["xcentroid_win"] = "x_centroid_win"
        name_map["ycentroid_win"] = "y_centroid_win"
        name_map["area"] = "segment_area"
        name_map["semimajor_sigma"] = "semimajor"
        name_map["semiminor_sigma"] = "semiminor"
        name_map["orientation"] = "orientation_pix"
        name_map["segment_fluxerr"] = "segment_flux_err"
        name_map["kron_fluxerr"] = "kron_flux_err"

        # set the source properties as attributes of this instance
        for name in photutils_names:
            new_name = name_map.get(name, name)
            value = getattr(segm_cat, name)

            # handle any unit conversions
            if new_name in (
                "x_centroid",
                "y_centroid",
                "x_centroid_win",
                "y_centroid_win",
            ):
                value *= u.pix

            # pix**2 -> arcsec**2
            if new_name == "segment_area":
                value = value.value * self.pixel_area.to(u.arcsec**2)

            # pix -> arcsec
            if new_name in ("semimajor", "semiminor", "fwhm", "kron_radius"):
                value = value.value * self.pixel_scale

            # 1 / pix**2 -> 1 / arcsec**-2
            if new_name in ("cxx", "cxy", "cyy"):
                value = value.value / self.pixel_area.to(u.arcsec**2)

            # remove dimensionless units
            if new_name == "ellipticity":
                value = value.value

            # change the photutils dtypes
            if new_name not in ("sky_centroid", "sky_centroid_win"):
                if np.issubdtype(value.dtype, np.integer):
                    value = value.astype(np.int32)
                elif np.issubdtype(value.dtype, np.floating):
                    value = value.astype(np.float32)

            # split the sky_centroid values into separate RA and Dec
            # values
            if new_name == "sky_centroid":
                self.ra_centroid = value.ra
                self.dec_centroid = value.dec
                self.names.extend(["ra_centroid", "dec_centroid"])
            elif new_name == "sky_centroid_win":
                self.ra_centroid_win = value.ra
                self.dec_centroid_win = value.dec
                self.names.extend(["ra_centroid_win", "dec_centroid_win"])
            else:
                setattr(self, new_name, value)
                self.names.append(new_name)

    def add_placeholders(self):
        """
        Add placeholder attributes to the class instance.
        """
        # pixel columns
        pix_columns = (
            "x_centroid_err",
            "y_centroid_err",
            "x_centroid_win_err",
            "y_centroid_win_err",
        )
        pix_value = np.zeros(self.source_cat.nlabels, dtype=np.float32) * u.pix
        for name in pix_columns:
            if not hasattr(self, name):
                setattr(self, name, pix_value)

        # sky columns
        sky_columns = (
            "ra_centroid_err",
            "dec_centroid_err",
            "ra_centroid_win_err",
            "dec_centroid_win_err",
        )
        sky_value = np.zeros(self.source_cat.nlabels, dtype=np.float32) * u.arcsec
        for name in sky_columns:
            if not hasattr(self, name):
                setattr(self, name, sky_value)

        for name in [*pix_columns, *sky_columns]:
            self.names.append(name)

    @lazyproperty
    def orientation_sky(self):
        """
        The position angle of the source major axis in degrees measured
        East of North.
        """
        return ((180.0 * u.deg) - self.wcs_angle + self.orientation_pix).astype(
            np.float32
        )

    @lazyproperty
    def _kron_abmag(self):
        return self.convert_flux_to_abmag(self.kron_flux, self.kron_flux_err)

    @lazyproperty
    def kron_abmag(self):
        """
        The Kron magnitude in AB magnitudes.
        """
        return self._kron_abmag[0]

    @lazyproperty
    def kron_abmag_err(self):
        """
        The Kron magnitude error in AB magnitudes.
        """
        return self._kron_abmag[1]

    @lazyproperty
    def fluxfrac_radius_50(self):
        """
        The radius (in arcsec) at which the flux fraction is 50%.
        """
        value = self.source_cat.fluxfrac_radius(0.5)
        return (value.value * self.pixel_scale).astype(np.float32)
